@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2012 Intel Corporation.
+ * Copyright © 2008-2014 Intel Corporation.
  * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
@@ -14,16 +14,10 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to:
- *
- *   Free Software Foundation, Inc.
- *   51 Franklin Street, Fifth Floor,
- *   Boston, MA 02110-1301 USA
  */
 
-#include <netdb.h>
+#include <config.h>
+
 #include <unistd.h>
 #include <fcntl.h>
 #include <time.h>
@@ -33,9 +27,6 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <sys/types.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
-#include <netinet/tcp.h>
 #include <stdarg.h>
 
 #include "openconnect-internal.h"
@@ -68,26 +59,6 @@ static struct pkt dpd_resp_pkt = {
 	.hdr = { 'S', 'T', 'F', 1, 0, 0, AC_PKT_DPD_RESP, 0 },
 };
 
-static int  __attribute__ ((format (printf, 3, 4)))
-    buf_append(char *buf, int len, const char *fmt, ...)
-{
-	int start = strlen(buf);
-	int ret;
-	va_list args;
-
-	if (start >= len)
-		return 0;
-
-	va_start(args, fmt);
-	ret = vsnprintf(buf + start, len - start, fmt, args);
-	va_end(args);
-
-	if (ret > len)
-		ret = len;
-
-	return ret;
-}
-
 /* Calculate MTU to request. Old servers simply use the X-CSTP-MTU: header,
  * which represents the tunnel MTU, while new servers do calculations on the
  * X-CSTP-Base-MTU: header which represents the cleartext MTU between client
@@ -116,7 +87,7 @@ static void calculate_mtu(struct openconnect_info *vpninfo, int *base_mtu, int *
 
 		if (!getsockopt(vpninfo->ssl_fd, IPPROTO_TCP, TCP_INFO,
 				&ti, &ti_size)) {
-			vpn_progress(vpninfo, PRG_TRACE,
+			vpn_progress(vpninfo, PRG_DEBUG,
 				     _("TCP_INFO rcv mss %d, snd mss %d, adv mss %d, pmtu %d\n"),
 				     ti.tcpi_rcv_mss, ti.tcpi_snd_mss, ti.tcpi_advmss, ti.tcpi_pmtu);
 			if (!*base_mtu)
@@ -136,7 +107,7 @@ static void calculate_mtu(struct openconnect_info *vpninfo, int *base_mtu, int *
 		socklen_t mss_size = sizeof(mss);
 		if (!getsockopt(vpninfo->ssl_fd, IPPROTO_TCP, TCP_MAXSEG,
 				&mss, &mss_size)) {
-			vpn_progress(vpninfo, PRG_TRACE, _("TCP_MAXSEG %d\n"), mss);
+			vpn_progress(vpninfo, PRG_DEBUG, _("TCP_MAXSEG %d\n"), mss);
 			*mtu = mss - 13;
 		}
 	}
@@ -149,83 +120,108 @@ static void calculate_mtu(struct openconnect_info *vpninfo, int *base_mtu, int *
 		*mtu = 1280;
 }
 
+void cstp_free_splits(struct openconnect_info *vpninfo)
+{
+	struct oc_split_include *inc;
+
+	for (inc = vpninfo->ip_info.split_includes; inc; ) {
+		struct oc_split_include *next = inc->next;
+		free(inc);
+		inc = next;
+	}
+	for (inc = vpninfo->ip_info.split_excludes; inc; ) {
+		struct oc_split_include *next = inc->next;
+		free(inc);
+		inc = next;
+	}
+	for (inc = vpninfo->ip_info.split_dns; inc; ) {
+		struct oc_split_include *next = inc->next;
+		free(inc);
+		inc = next;
+	}
+	vpninfo->ip_info.split_dns = vpninfo->ip_info.split_includes =
+		vpninfo->ip_info.split_excludes = NULL;
+}
+
+/* if DTLS 1.2 is supported */
+#if defined(DTLS_GNUTLS) && GNUTLS_VERSION_NUMBER >= 0x030200
+# define DEFAULT_CIPHER_LIST "OC-DTLS1_2-AES256-GCM:OC-DTLS1_2-AES128-GCM:AES256-SHA:AES128-SHA:DES-CBC3-SHA:DES-CBC-SHA"
+#else
+# define DEFAULT_CIPHER_LIST "AES256-SHA:AES128-SHA:DES-CBC3-SHA:DES-CBC-SHA"
+#endif
+
 static int start_cstp_connection(struct openconnect_info *vpninfo)
 {
+	struct oc_text_buf *reqbuf;
 	char buf[65536];
 	int i;
+	int dtls_secret_set = 0;
 	int retried = 0, sessid_found = 0;
-	struct vpn_option **next_dtls_option = &vpninfo->dtls_options;
-	struct vpn_option **next_cstp_option = &vpninfo->cstp_options;
-	struct vpn_option *old_cstp_opts = vpninfo->cstp_options;
-	struct vpn_option *old_dtls_opts = vpninfo->dtls_options;
-	const char *old_addr = vpninfo->vpn_addr;
-	const char *old_netmask = vpninfo->vpn_netmask;
-	const char *old_addr6 = vpninfo->vpn_addr6;
-	const char *old_netmask6 = vpninfo->vpn_netmask6;
-	struct split_include *inc;
+	struct oc_vpn_option **next_dtls_option = &vpninfo->dtls_options;
+	struct oc_vpn_option **next_cstp_option = &vpninfo->cstp_options;
+	struct oc_vpn_option *old_cstp_opts = vpninfo->cstp_options;
+	struct oc_vpn_option *old_dtls_opts = vpninfo->dtls_options;
+	const char *old_addr = vpninfo->ip_info.addr;
+	const char *old_netmask = vpninfo->ip_info.netmask;
+	const char *old_addr6 = vpninfo->ip_info.addr6;
+	const char *old_netmask6 = vpninfo->ip_info.netmask6;
 	int base_mtu, mtu;
 
 	/* Clear old options which will be overwritten */
-	vpninfo->vpn_addr = vpninfo->vpn_netmask = NULL;
-	vpninfo->vpn_addr6 = vpninfo->vpn_netmask6 = NULL;
+	vpninfo->ip_info.addr = vpninfo->ip_info.netmask = NULL;
+	vpninfo->ip_info.addr6 = vpninfo->ip_info.netmask6 = NULL;
 	vpninfo->cstp_options = vpninfo->dtls_options = NULL;
-	vpninfo->vpn_domain = vpninfo->vpn_proxy_pac = NULL;
+	vpninfo->ip_info.domain = vpninfo->ip_info.proxy_pac = NULL;
 	vpninfo->banner = NULL;
 
 	for (i = 0; i < 3; i++)
-		vpninfo->vpn_dns[i] = vpninfo->vpn_nbns[i] = NULL;
-
-	for (inc = vpninfo->split_includes; inc; ) {
-		struct split_include *next = inc->next;
-		free(inc);
-		inc = next;
-	}
-	for (inc = vpninfo->split_excludes; inc; ) {
-		struct split_include *next = inc->next;
-		free(inc);
-		inc = next;
-	}
-	for (inc = vpninfo->split_dns; inc; ) {
-		struct split_include *next = inc->next;
-		free(inc);
-		inc = next;
-	}
-	vpninfo->split_dns = vpninfo->split_includes = vpninfo->split_excludes = NULL;
-
-	/* Create (new) random master key for DTLS connection, if needed */
-	if (vpninfo->dtls_times.last_rekey + vpninfo->dtls_times.rekey <
-	    time(NULL) + 300 &&
-	    openconnect_random(vpninfo->dtls_secret, sizeof(vpninfo->dtls_secret))) {
-		fprintf(stderr, _("Failed to initialise DTLS secret\n"));
-		exit(1);
-	}
+		vpninfo->ip_info.dns[i] = vpninfo->ip_info.nbns[i] = NULL;
+	cstp_free_splits(vpninfo);
 
  retry:
 	calculate_mtu(vpninfo, &base_mtu, &mtu);
 
-	buf[0] = 0;
-	buf_append(buf, sizeof(buf), "CONNECT /CSCOSSLC/tunnel HTTP/1.1\r\n");
-	buf_append(buf, sizeof(buf), "Host: %s\r\n", vpninfo->hostname);
-	buf_append(buf, sizeof(buf), "User-Agent: %s\r\n", vpninfo->useragent);
-	buf_append(buf, sizeof(buf), "Cookie: webvpn=%s\r\n", vpninfo->cookie);
-	buf_append(buf, sizeof(buf), "X-CSTP-Version: 1\r\n");
-	buf_append(buf, sizeof(buf), "X-CSTP-Hostname: %s\r\n", vpninfo->localname);
+	reqbuf = buf_alloc();
+	buf_append(reqbuf, "CONNECT /CSCOSSLC/tunnel HTTP/1.1\r\n");
+	buf_append(reqbuf, "Host: %s\r\n", vpninfo->hostname);
+	buf_append(reqbuf, "User-Agent: %s\r\n", vpninfo->useragent);
+	buf_append(reqbuf, "Cookie: webvpn=%s\r\n", vpninfo->cookie);
+	buf_append(reqbuf, "X-CSTP-Version: 1\r\n");
+	buf_append(reqbuf, "X-CSTP-Hostname: %s\r\n", vpninfo->localname);
 	if (vpninfo->deflate && i < sizeof(buf))
-		buf_append(buf, sizeof(buf), "X-CSTP-Accept-Encoding: deflate;q=1.0\r\n");
+		buf_append(reqbuf, "X-CSTP-Accept-Encoding: deflate;q=1.0\r\n");
 	if (base_mtu)
-		buf_append(buf, sizeof(buf), "X-CSTP-Base-MTU: %d\r\n", base_mtu);
-	buf_append(buf, sizeof(buf), "X-CSTP-MTU: %d\r\n", mtu);
-	buf_append(buf, sizeof(buf), "X-CSTP-Address-Type: %s\r\n",
+		buf_append(reqbuf, "X-CSTP-Base-MTU: %d\r\n", base_mtu);
+	buf_append(reqbuf, "X-CSTP-MTU: %d\r\n", mtu);
+	buf_append(reqbuf, "X-CSTP-Address-Type: %s\r\n",
 			       vpninfo->disable_ipv6 ? "IPv4" : "IPv6,IPv4");
-	buf_append(buf, sizeof(buf), "X-DTLS-Master-Secret: ");
-	for (i = 0; i < sizeof(vpninfo->dtls_secret); i++)
-		buf_append(buf, sizeof(buf), "%02X", vpninfo->dtls_secret[i]);
-	buf_append(buf, sizeof(buf), "\r\nX-DTLS-CipherSuite: %s\r\n\r\n",
-			       vpninfo->dtls_ciphers ? : "AES256-SHA:AES128-SHA:DES-CBC3-SHA:DES-CBC-SHA");
+	if (!vpninfo->disable_ipv6)
+		buf_append(reqbuf, "X-CSTP-Full-IPv6-Capability: true\r\n");
+	buf_append(reqbuf, "X-DTLS-Master-Secret: ");
+	for (i = 0; i < sizeof(vpninfo->dtls_secret); i++) {
+		buf_append(reqbuf, "%02X", vpninfo->dtls_secret[i]);
+		dtls_secret_set |= vpninfo->dtls_secret[i];
+	}
+	if (!dtls_secret_set) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("CRITICAL ERROR: DTLS master secret is uninitialised. Please report this.\n"));
+		return -EINVAL;
+	}
 
-	openconnect_SSL_write(vpninfo, buf, strlen(buf));
+	buf_append(reqbuf, "\r\nX-DTLS-CipherSuite: %s\r\n\r\n",
+			       vpninfo->dtls_ciphers ? : DEFAULT_CIPHER_LIST);
 
-	if ((i = openconnect_SSL_gets(vpninfo, buf, 65536) < 0)) {
+	if (buf_error(reqbuf)) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Error creating HTTPS CONNECT request\n"));
+		return buf_free(reqbuf);
+	}
+	vpninfo->ssl_write(vpninfo, reqbuf->data, reqbuf->pos);
+	buf_free(reqbuf);
+
+	/* FIXME: Use process_http_response() instead of reimplementing it. It has
+	   a header callback function, and can cope with CONNECT requests. */
+	if ((i = vpninfo->ssl_gets(vpninfo, buf, 65536)) < 0) {
 		if (i == -EINTR)
 			return i;
 		vpn_progress(vpninfo, PRG_ERR,
@@ -238,7 +234,7 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("Failed to open HTTPS connection to %s\n"),
 					     vpninfo->hostname);
-				exit(1);
+				return -EIO;
 			}
 			goto retry;
 		}
@@ -249,7 +245,7 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 		if (!strncmp(buf, "HTTP/1.1 503 ", 13)) {
 			/* "Service Unavailable. Why? */
 			const char *reason = "<unknown>";
-			while ((i = openconnect_SSL_gets(vpninfo, buf, sizeof(buf)))) {
+			while ((i = vpninfo->ssl_gets(vpninfo, buf, sizeof(buf)))) {
 				if (!strncmp(buf, "X-Reason: ", 10)) {
 					reason = buf + 10;
 					break;
@@ -264,7 +260,7 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 			     _("Got inappropriate HTTP CONNECT response: %s\n"),
 			     buf);
 		if (!strncmp(buf, "HTTP/1.1 401 ", 13))
-			exit(2);
+			return -EPERM;
 		return -EINVAL;
 	}
 
@@ -274,8 +270,8 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	vpninfo->deflate = 0;
 	mtu = 0;
 
-	while ((i = openconnect_SSL_gets(vpninfo, buf, sizeof(buf)))) {
-		struct vpn_option *new_option;
+	while ((i = vpninfo->ssl_gets(vpninfo, buf, sizeof(buf)))) {
+		struct oc_vpn_option *new_option;
 		char *colon;
 
 		if (i < 0)
@@ -313,9 +309,9 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 
 		/* This contains the whole document, including the webvpn cookie. */
 		if (!strcasecmp(buf, "X-CSTP-Post-Auth-XML"))
-			vpn_progress(vpninfo, PRG_TRACE, "%s: %s\n", buf, _("<elided>"));
+			vpn_progress(vpninfo, PRG_DEBUG, "%s: %s\n", buf, _("<elided>"));
 		else
-			vpn_progress(vpninfo, PRG_TRACE, "%s: %s\n", buf, colon);
+			vpn_progress(vpninfo, PRG_DEBUG, "%s: %s\n", buf, colon);
 
 		if (!strncmp(buf, "X-DTLS-", 7)) {
 			*next_dtls_option = new_option;
@@ -336,7 +332,6 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 				for (i = 0; i < 64; i += 2)
 					vpninfo->dtls_session_id[i/2] = unhex(colon + i);
 				sessid_found = 1;
-				time(&vpninfo->dtls_times.last_rekey);
 			}
 			continue;
 		}
@@ -353,6 +348,13 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 				vpninfo->ssl_times.dpd = j;
 		} else if (!strcmp(buf + 7, "Rekey-Time")) {
 			vpninfo->ssl_times.rekey = atol(colon);
+		} else if (!strcmp(buf + 7, "Rekey-Method")) {
+			if (!strcmp(colon, "new-tunnel"))
+				vpninfo->ssl_times.rekey_method = REKEY_TUNNEL;
+			else if (!strcmp(colon, "ssl"))
+				vpninfo->ssl_times.rekey_method = REKEY_SSL;
+			else
+				vpninfo->ssl_times.rekey_method = REKEY_NONE;
 		} else if (!strcmp(buf + 7, "Content-Encoding")) {
 			if (!strcmp(colon, "deflate"))
 				vpninfo->deflate = 1;
@@ -366,61 +368,63 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 			int cstpmtu = atol(colon);
 			if (cstpmtu > mtu)
 				mtu = cstpmtu;
+		} else if (!strcmp(buf + 7, "Address-IP6")) {
+			vpninfo->ip_info.netmask6 = new_option->value;
 		} else if (!strcmp(buf + 7, "Address")) {
 			if (strchr(new_option->value, ':')) {
 				if (!vpninfo->disable_ipv6)
-					vpninfo->vpn_addr6 = new_option->value;
+					vpninfo->ip_info.addr6 = new_option->value;
 			} else
-				vpninfo->vpn_addr = new_option->value;
+				vpninfo->ip_info.addr = new_option->value;
 		} else if (!strcmp(buf + 7, "Netmask")) {
 			if (strchr(new_option->value, ':')) {
 				if (!vpninfo->disable_ipv6)
-					vpninfo->vpn_netmask6 = new_option->value;
+					vpninfo->ip_info.netmask6 = new_option->value;
 			} else
-				vpninfo->vpn_netmask = new_option->value;
+				vpninfo->ip_info.netmask = new_option->value;
 		} else if (!strcmp(buf + 7, "DNS")) {
 			int j;
 			for (j = 0; j < 3; j++) {
-				if (!vpninfo->vpn_dns[j]) {
-					vpninfo->vpn_dns[j] = new_option->value;
+				if (!vpninfo->ip_info.dns[j]) {
+					vpninfo->ip_info.dns[j] = new_option->value;
 					break;
 				}
 			}
 		} else if (!strcmp(buf + 7, "NBNS")) {
 			int j;
 			for (j = 0; j < 3; j++) {
-				if (!vpninfo->vpn_nbns[j]) {
-					vpninfo->vpn_nbns[j] = new_option->value;
+				if (!vpninfo->ip_info.nbns[j]) {
+					vpninfo->ip_info.nbns[j] = new_option->value;
 					break;
 				}
 			}
 		} else if (!strcmp(buf + 7, "Default-Domain")) {
-			vpninfo->vpn_domain = new_option->value;
+			vpninfo->ip_info.domain = new_option->value;
 		} else if (!strcmp(buf + 7, "MSIE-Proxy-PAC-URL")) {
-			vpninfo->vpn_proxy_pac = new_option->value;
+			vpninfo->ip_info.proxy_pac = new_option->value;
 		} else if (!strcmp(buf + 7, "Banner")) {
 			vpninfo->banner = new_option->value;
 		} else if (!strcmp(buf + 7, "Split-DNS")) {
-			struct split_include *dns = malloc(sizeof(*dns));
+			struct oc_split_include *dns = malloc(sizeof(*dns));
 			if (!dns)
 				continue;
 			dns->route = new_option->value;
-			dns->next = vpninfo->split_dns;
-			vpninfo->split_dns = dns;
-		} else if (!strcmp(buf + 7, "Split-Include")) {
-			struct split_include *inc = malloc(sizeof(*inc));
+			dns->next = vpninfo->ip_info.split_dns;
+			vpninfo->ip_info.split_dns = dns;
+		} else if (!strcmp(buf + 7, "Split-Include") || !strcmp(buf + 7, "Split-Include-IP6")) {
+			struct oc_split_include *inc = malloc(sizeof(*inc));
 			if (!inc)
 				continue;
 			inc->route = new_option->value;
-			inc->next = vpninfo->split_includes;
-			vpninfo->split_includes = inc;
-		} else if (!strcmp(buf + 7, "Split-Exclude")) {
-			struct split_include *exc = malloc(sizeof(*exc));
+			inc->next = vpninfo->ip_info.split_includes;
+			vpninfo->ip_info.split_includes = inc;
+		} else if (!strcmp(buf + 7, "Split-Exclude") || !strcmp(buf + 7, "Split-Exclude-IP6")) {
+			struct oc_split_include *exc = malloc(sizeof(*exc));
 			if (!exc)
 				continue;
 			exc->route = new_option->value;
-			exc->next = vpninfo->split_excludes;
-			vpninfo->split_excludes = exc;
+			exc->next = vpninfo->ip_info.split_excludes;
+			vpninfo->ip_info.split_excludes = exc;
 		}
 	}
 
@@ -429,55 +433,55 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 			     _("No MTU received. Aborting\n"));
 		return -EINVAL;
 	}
-	vpninfo->actual_mtu = mtu;
+	vpninfo->ip_info.mtu = mtu;
 
-	if (!vpninfo->vpn_addr && !vpninfo->vpn_addr6) {
+	if (!vpninfo->ip_info.addr && !vpninfo->ip_info.addr6) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("No IP address received. Aborting\n"));
 		return -EINVAL;
 	}
 	if (old_addr) {
-		if (strcmp(old_addr, vpninfo->vpn_addr)) {
+		if (strcmp(old_addr, vpninfo->ip_info.addr)) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Reconnect gave different Legacy IP address (%s != %s)\n"),
-				     vpninfo->vpn_addr, old_addr);
+				     vpninfo->ip_info.addr, old_addr);
 			return -EINVAL;
 		}
 	}
 	if (old_netmask) {
-		if (strcmp(old_netmask, vpninfo->vpn_netmask)) {
+		if (strcmp(old_netmask, vpninfo->ip_info.netmask)) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Reconnect gave different Legacy IP netmask (%s != %s)\n"),
-				     vpninfo->vpn_netmask, old_netmask);
+				     vpninfo->ip_info.netmask, old_netmask);
 			return -EINVAL;
 		}
 	}
 	if (old_addr6) {
-		if (strcmp(old_addr6, vpninfo->vpn_addr6)) {
+		if (strcmp(old_addr6, vpninfo->ip_info.addr6)) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Reconnect gave different IPv6 address (%s != %s)\n"),
-				     vpninfo->vpn_addr6, old_addr6);
+				     vpninfo->ip_info.addr6, old_addr6);
 			return -EINVAL;
 		}
 	}
 	if (old_netmask6) {
-		if (strcmp(old_netmask6, vpninfo->vpn_netmask6)) {
+		if (strcmp(old_netmask6, vpninfo->ip_info.netmask6)) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Reconnect gave different IPv6 netmask (%s != %s)\n"),
-				     vpninfo->vpn_netmask6, old_netmask6);
+				     vpninfo->ip_info.netmask6, old_netmask6);
 			return -EINVAL;
 		}
 	}
 
 	while (old_dtls_opts) {
-		struct vpn_option *tmp = old_dtls_opts;
+		struct oc_vpn_option *tmp = old_dtls_opts;
 		old_dtls_opts = old_dtls_opts->next;
 		free(tmp->value);
 		free(tmp->option);
 		free(tmp);
 	}
 	while (old_cstp_opts) {
-		struct vpn_option *tmp = old_cstp_opts;
+		struct oc_vpn_option *tmp = old_cstp_opts;
 		old_cstp_opts = old_cstp_opts->next;
 		free(tmp->value);
 		free(tmp->option);
@@ -486,25 +490,38 @@ static int start_cstp_connection(struct openconnect_info *vpninfo)
 	vpn_progress(vpninfo, PRG_INFO, _("CSTP connected. DPD %d, Keepalive %d\n"),
 		     vpninfo->ssl_times.dpd, vpninfo->ssl_times.keepalive);
 
-	if (vpninfo->select_nfds <= vpninfo->ssl_fd)
-		vpninfo->select_nfds = vpninfo->ssl_fd + 1;
 
-	FD_SET(vpninfo->ssl_fd, &vpninfo->select_rfds);
-	FD_SET(vpninfo->ssl_fd, &vpninfo->select_efds);
+	monitor_fd_new(vpninfo, ssl);
+
+	monitor_read_fd(vpninfo, ssl);
+	monitor_except_fd(vpninfo, ssl);
 
 	if (!sessid_found)
 		vpninfo->dtls_attempt_period = 0;
 
+	if (vpninfo->ssl_times.rekey <= 0)
+		vpninfo->ssl_times.rekey_method = REKEY_NONE;
+
 	vpninfo->ssl_times.last_rekey = vpninfo->ssl_times.last_rx =
 		vpninfo->ssl_times.last_tx = time(NULL);
 	return 0;
 }
 
 
-int make_cstp_connection(struct openconnect_info *vpninfo)
+int openconnect_make_cstp_connection(struct openconnect_info *vpninfo)
 {
 	int ret;
 
+	/* This needs to be done before openconnect_setup_dtls() because it's
+	   sent with the CSTP CONNECT handshake. Even if we don't end up doing
+	   DTLS. */
+	if (vpninfo->dtls_state == DTLS_NOSECRET) {
+		if (openconnect_random(vpninfo->dtls_secret, sizeof(vpninfo->dtls_secret)))
+			return -EINVAL;
+		/* The application will later call openconnect_setup_dtls() */
+		vpninfo->dtls_state = DTLS_DISABLED;
+	}
+
 	ret = openconnect_open_https(vpninfo);
 	if (ret)
 		return ret;
@@ -536,10 +553,13 @@ int make_cstp_connection(struct openconnect_info *vpninfo)
 		}
 	}
 
-	return start_cstp_connection(vpninfo);
+	ret = start_cstp_connection(vpninfo);
+	if (ret < 0)
+		openconnect_close_https(vpninfo, 0);
+	return ret;
 }
 
-int cstp_reconnect(struct openconnect_info *vpninfo)
+static int cstp_reconnect(struct openconnect_info *vpninfo)
 {
 	int ret;
 	int timeout;
@@ -560,15 +580,22 @@ int cstp_reconnect(struct openconnect_info *vpninfo)
 	timeout = vpninfo->reconnect_timeout;
 	interval = vpninfo->reconnect_interval;
 
-	while ((ret = make_cstp_connection(vpninfo))) {
+	while ((ret = openconnect_make_cstp_connection(vpninfo))) {
 		if (timeout <= 0)
 			return ret;
+		if (ret == -EPERM) {
+			vpn_progress(vpninfo, PRG_ERR,
+				     _("Cookie is no longer valid, ending session\n"));
+			return ret;
+		}
 		vpn_progress(vpninfo, PRG_INFO,
 			     _("sleep %ds, remaining timeout %ds\n"),
 			     interval, timeout);
-		sleep(interval);
-		if (killed)
-			return 1;
+		poll_cmd_fd(vpninfo, interval);
+		if (vpninfo->got_cancel_cmd)
+			return -EINTR;
+		if (vpninfo->got_pause_cmd)
+			return 0;
 		timeout -= interval;
 		interval += vpninfo->reconnect_interval;
 		if (interval > RECONNECT_INTERVAL_MAX)
@@ -581,7 +608,7 @@ int cstp_reconnect(struct openconnect_info *vpninfo)
 static int inflate_and_queue_packet(struct openconnect_info *vpninfo,
 				    unsigned char *buf, int len)
 {
-	struct pkt *new = malloc(sizeof(struct pkt) + vpninfo->actual_mtu);
+	struct pkt *new = malloc(sizeof(struct pkt) + vpninfo->ip_info.mtu);
 	uint32_t pkt_sum;
 
 	if (!new)
@@ -593,7 +620,7 @@ static int inflate_and_queue_packet(struct openconnect_info *vpninfo,
 	vpninfo->inflate_strm.avail_in = len - 4;
 
 	vpninfo->inflate_strm.next_out = new->data;
-	vpninfo->inflate_strm.avail_out = vpninfo->actual_mtu;
+	vpninfo->inflate_strm.avail_out = vpninfo->ip_info.mtu;
 	vpninfo->inflate_strm.total_out = 0;
 
 	if (inflate(&vpninfo->inflate_strm, Z_SYNC_FLUSH)) {
@@ -653,7 +680,7 @@ static int cstp_write(struct openconnect_info *vpninfo, void *buf, int buflen)
 	case SSL_ERROR_WANT_WRITE:
 		/* Waiting for the socket to become writable -- it's
 		   probably stalled, and/or the buffers are full */
-		FD_SET(vpninfo->ssl_fd, &vpninfo->select_wfds);
+		monitor_write_fd(vpninfo, ssl);
 	case SSL_ERROR_WANT_READ:
 		return 0;
 
@@ -693,7 +720,7 @@ static int cstp_write(struct openconnect_info *vpninfo, void *buf, int buflen)
 		if (gnutls_record_get_direction(vpninfo->https_sess)) {
 			/* Waiting for the socket to become writable -- it's
 			   probably stalled, and/or the buffers are full */
-			FD_SET(vpninfo->ssl_fd, &vpninfo->select_wfds);
+			monitor_write_fd(vpninfo, ssl);
 		}
 		return 0;
 	}
@@ -709,6 +736,9 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	int len, ret;
 	int work_done = 0;
 
+	if (vpninfo->ssl_fd == -1)
+		goto do_reconnect;
+
 	/* FIXME: The poll() handling here is fairly simplistic. Actually,
 	   if the SSL connection stalls it could return a WANT_WRITE error
 	   on _either_ of the SSL_read() or SSL_write() calls. In that case,
@@ -736,18 +766,18 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 		vpninfo->ssl_times.last_rx = time(NULL);
 		switch (buf[6]) {
 		case AC_PKT_DPD_OUT:
-			vpn_progress(vpninfo, PRG_TRACE,
+			vpn_progress(vpninfo, PRG_DEBUG,
 				     _("Got CSTP DPD request\n"));
 			vpninfo->owe_ssl_dpd_response = 1;
 			continue;
 
 		case AC_PKT_DPD_RESP:
-			vpn_progress(vpninfo, PRG_TRACE,
+			vpn_progress(vpninfo, PRG_DEBUG,
 				     _("Got CSTP DPD response\n"));
 			continue;
 
 		case AC_PKT_KEEPALIVE:
-			vpn_progress(vpninfo, PRG_TRACE,
+			vpn_progress(vpninfo, PRG_DEBUG,
 				     _("Got CSTP Keepalive\n"));
 			continue;
 
@@ -771,7 +801,7 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 				     _("Received server disconnect: %02x '%s'\n"),
 				     buf[8], buf + 9);
 			vpninfo->quit_reason = "Server request";
-			return 1;
+			return -EPIPE;
 		}
 		case AC_PKT_COMPRESSED:
 			if (!vpninfo->deflate) {
@@ -786,7 +816,7 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 		case AC_PKT_TERM_SERVER:
 			vpn_progress(vpninfo, PRG_ERR, _("received server terminate packet\n"));
 			vpninfo->quit_reason = "Server request";
-			return 1;
+			return -EPIPE;
 		}
 
 	unknown_pkt:
@@ -807,7 +837,7 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	if (vpninfo->current_ssl_pkt) {
 	handle_outgoing:
 		vpninfo->ssl_times.last_tx = time(NULL);
-		FD_CLR(vpninfo->ssl_fd, &vpninfo->select_wfds);
+		unmonitor_write_fd(vpninfo, ssl);
 
 		ret = cstp_write(vpninfo,
 				 vpninfo->current_ssl_pkt->hdr,
@@ -862,7 +892,18 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 		/* Not that this will ever happen; we don't even process
 		   the setting when we're asked for it. */
 		vpn_progress(vpninfo, PRG_INFO, _("CSTP rekey due\n"));
-		goto do_reconnect;
+		if (vpninfo->ssl_times.rekey_method == REKEY_TUNNEL)
+			goto do_reconnect;
+		else if (vpninfo->ssl_times.rekey_method == REKEY_SSL) {
+			ret = cstp_handshake(vpninfo, 0);
+			if (ret) {
+				/* if we failed rehandshake try establishing a new-tunnel instead of failing */
+				vpn_progress(vpninfo, PRG_ERR, _("Rehandshake failed; attempting new-tunnel\n"));
+				goto do_reconnect;
+			}
+
+			goto do_dtls_reconnect;
+		}
 		break;
 
 	case KA_DPD_DEAD:
@@ -870,17 +911,25 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("CSTP Dead Peer Detection detected dead peer!\n"));
 	do_reconnect:
-		if (cstp_reconnect(vpninfo)) {
+		ret = cstp_reconnect(vpninfo);
+		if (ret) {
 			vpn_progress(vpninfo, PRG_ERR, _("Reconnect failed\n"));
 			vpninfo->quit_reason = "CSTP reconnect failed";
-			return 1;
+			return ret;
 		}
-		/* I think we can leave DTLS to its own devices; when we reconnect
-		   with the same master secret, we do seem to get the same sessid */
+
+	do_dtls_reconnect:
+		/* succeeded, let's rekey DTLS, if it is not rekeying
+		 * itself. */
+		if (vpninfo->dtls_state != DTLS_DISABLED &&
+		    vpninfo->dtls_times.rekey_method == REKEY_NONE) {
+			dtls_reconnect(vpninfo);
+		}
+
 		return 1;
 
 	case KA_DPD:
-		vpn_progress(vpninfo, PRG_TRACE, _("Send CSTP DPD\n"));
+		vpn_progress(vpninfo, PRG_DEBUG, _("Send CSTP DPD\n"));
 
 		vpninfo->current_ssl_pkt = &dpd_pkt;
 		goto handle_outgoing;
@@ -888,10 +937,10 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	case KA_KEEPALIVE:
 		/* No need to send an explicit keepalive
 		   if we have real data to send */
-		if (vpninfo->dtls_fd == -1 && vpninfo->outgoing_queue)
+		if (vpninfo->dtls_state != DTLS_CONNECTED && vpninfo->outgoing_queue)
 			break;
 
-		vpn_progress(vpninfo, PRG_TRACE, _("Send CSTP Keepalive\n"));
+		vpn_progress(vpninfo, PRG_DEBUG, _("Send CSTP Keepalive\n"));
 
 		vpninfo->current_ssl_pkt = &keepalive_pkt;
 		goto handle_outgoing;
@@ -901,7 +950,7 @@ int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	}
 
 	/* Service outgoing packet queue, if no DTLS */
-	while (vpninfo->dtls_fd == -1 && vpninfo->outgoing_queue) {
+	while (vpninfo->dtls_state != DTLS_CONNECTED && vpninfo->outgoing_queue) {
 		struct pkt *this = vpninfo->outgoing_queue;
 		vpninfo->outgoing_queue = this->next;
 		vpninfo->outgoing_qlen--;
