@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2012 Intel Corporation.
+ * Copyright © 2008-2014 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -13,21 +13,13 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to:
- *
- *   Free Software Foundation, Inc.
- *   51 Franklin Street, Fifth Floor,
- *   Boston, MA 02110-1301 USA
  */
 
+#include <config.h>
+
 #include <errno.h>
 #include <sys/types.h>
-#include <sys/socket.h>
-#include <netdb.h>
 #include <unistd.h>
-#include <netinet/in.h>
 #include <fcntl.h>
 #include <string.h>
 #include <errno.h>
@@ -109,6 +101,7 @@ int RAND_bytes(char *buf, int len)
 #if defined(DTLS_OPENSSL)
 #define DTLS_SEND SSL_write
 #define DTLS_RECV SSL_read
+#define DTLS_FREE SSL_free
 
 #ifdef HAVE_DTLS1_STOP_TIMER
 /* OpenSSL doesn't deliberately export this, but we need it to
@@ -209,33 +202,23 @@ static int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 
 	SSL_set_options(dtls_ssl, SSL_OP_CISCO_ANYCONNECT);
 
-	vpninfo->new_dtls_ssl = dtls_ssl;
+	vpninfo->dtls_ssl = dtls_ssl;
 
 	return 0;
 }
 
 int dtls_try_handshake(struct openconnect_info *vpninfo)
 {
-	int ret = SSL_do_handshake(vpninfo->new_dtls_ssl);
+	int ret = SSL_do_handshake(vpninfo->dtls_ssl);
 
 	if (ret == 1) {
-		vpn_progress(vpninfo, PRG_INFO, _("Established DTLS connection (using OpenSSL)\n"));
+		vpninfo->dtls_state = DTLS_CONNECTED;
+		vpn_progress(vpninfo, PRG_INFO,
+			     _("Established DTLS connection (using OpenSSL). Ciphersuite %s.\n"),
+			     vpninfo->dtls_cipher);
 
-		if (vpninfo->dtls_ssl) {
-			/* We are replacing an old connection */
-			SSL_free(vpninfo->dtls_ssl);
-			close(vpninfo->dtls_fd);
-			FD_CLR(vpninfo->dtls_fd, &vpninfo->select_rfds);
-			FD_CLR(vpninfo->dtls_fd, &vpninfo->select_wfds);
-			FD_CLR(vpninfo->dtls_fd, &vpninfo->select_efds);
-		}
-		vpninfo->dtls_ssl = vpninfo->new_dtls_ssl;
-		vpninfo->dtls_fd = vpninfo->new_dtls_fd;
-
-		vpninfo->new_dtls_ssl = NULL;
-		vpninfo->new_dtls_fd = -1;
-
-		vpninfo->dtls_times.last_rx = vpninfo->dtls_times.last_tx = time(NULL);
+		vpninfo->dtls_times.last_rekey = vpninfo->dtls_times.last_rx = 
+			vpninfo->dtls_times.last_tx = time(NULL);
 
 		/* From about 8.4.1(11) onwards, the ASA seems to get
 		   very unhappy if we resend ChangeCipherSpec messages
@@ -297,10 +280,10 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 		return 0;
 	}
 
-	ret = SSL_get_error(vpninfo->new_dtls_ssl, ret);
+	ret = SSL_get_error(vpninfo->dtls_ssl, ret);
 	if (ret == SSL_ERROR_WANT_WRITE || ret == SSL_ERROR_WANT_READ) {
 		static int badossl_bitched = 0;
-		if (time(NULL) < vpninfo->new_dtls_started + 5)
+		if (time(NULL) < vpninfo->new_dtls_started + 12)
 			return 0;
 		if (((OPENSSL_VERSION_NUMBER >= 0x100000b0L && OPENSSL_VERSION_NUMBER <= 0x100000c0L) || \
 		     (OPENSSL_VERSION_NUMBER >= 0x10001040L && OPENSSL_VERSION_NUMBER <= 0x10001060L) || \
@@ -310,57 +293,54 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 			vpn_progress(vpninfo, PRG_ERR, _("This is probably because your OpenSSL is broken\n"
 				"See http://rt.openssl.org/Ticket/Display.html?id=2984\n"));
 		} else {
-			vpn_progress(vpninfo, PRG_TRACE, _("DTLS handshake timed out\n"));
+			vpn_progress(vpninfo, PRG_DEBUG, _("DTLS handshake timed out\n"));
 		}
 	}
 
 	vpn_progress(vpninfo, PRG_ERR, _("DTLS handshake failed: %d\n"), ret);
 	openconnect_report_ssl_errors(vpninfo);
 
-	/* Kill the new (failed) connection... */
-	SSL_free(vpninfo->new_dtls_ssl);
-	FD_CLR(vpninfo->new_dtls_fd, &vpninfo->select_rfds);
-	FD_CLR(vpninfo->new_dtls_fd, &vpninfo->select_efds);
-	close(vpninfo->new_dtls_fd);
-	vpninfo->new_dtls_ssl = NULL;
-	vpninfo->new_dtls_fd = -1;
-
-	/* ... and kill the old one too. The only time there'll be a valid
-	   existing session is when it was a rekey, and in that case it's
-	   time for the old one to die. */
-	if (vpninfo->dtls_ssl) {
-		SSL_free(vpninfo->dtls_ssl);
-		close(vpninfo->dtls_fd);
-		FD_CLR(vpninfo->dtls_fd, &vpninfo->select_rfds);
-		FD_CLR(vpninfo->dtls_fd, &vpninfo->select_wfds);
-		FD_CLR(vpninfo->dtls_fd, &vpninfo->select_efds);
-		vpninfo->dtls_ssl = NULL;
-		vpninfo->dtls_fd = -1;
-	}
+	dtls_close(vpninfo);
 
+	vpninfo->dtls_state = DTLS_SLEEPING;
 	time(&vpninfo->new_dtls_started);
 	return -EINVAL;
 }
 
+void dtls_shutdown(struct openconnect_info *vpninfo)
+{
+	dtls_close(vpninfo);
+	SSL_CTX_free(vpninfo->dtls_ctx);
+	SSL_SESSION_free(vpninfo->dtls_session);
+}
+
 #elif defined(DTLS_GNUTLS)
 #include <gnutls/dtls.h>
 
 struct {
 	const char *name;
+	gnutls_protocol_t version;
 	gnutls_cipher_algorithm_t cipher;
 	gnutls_mac_algorithm_t mac;
 	const char *prio;
 } gnutls_dtls_ciphers[] = {
-	{ "AES128-SHA", GNUTLS_CIPHER_AES_128_CBC, GNUTLS_MAC_SHA1,
+	{ "AES128-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_AES_128_CBC, GNUTLS_MAC_SHA1,
 	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-128-CBC:+SHA1:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION" },
-	{ "AES256-SHA", GNUTLS_CIPHER_AES_256_CBC, GNUTLS_MAC_SHA1,
+	{ "AES256-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_AES_256_CBC, GNUTLS_MAC_SHA1,
 	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+AES-256-CBC:+SHA1:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION" },
-	{ "DES-CBC3-SHA", GNUTLS_CIPHER_3DES_CBC, GNUTLS_MAC_SHA1,
+	{ "DES-CBC3-SHA", GNUTLS_DTLS0_9, GNUTLS_CIPHER_3DES_CBC, GNUTLS_MAC_SHA1,
 	  "NONE:+VERS-DTLS0.9:+COMP-NULL:+3DES-CBC:+SHA1:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION" },
+#if GNUTLS_VERSION_NUMBER >= 0x030207 /* if DTLS 1.2 is supported (and a bug in gnutls is solved) */
+	{ "OC-DTLS1_2-AES128-GCM", GNUTLS_DTLS1_2, GNUTLS_CIPHER_AES_128_GCM, GNUTLS_MAC_AEAD,
+	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+AES-128-GCM:+AEAD:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION:+SIGN-ALL" },
+	{ "OC-DTLS1_2-AES256-GCM", GNUTLS_DTLS1_2, GNUTLS_CIPHER_AES_256_GCM, GNUTLS_MAC_AEAD,
+	  "NONE:+VERS-DTLS1.2:+COMP-NULL:+AES-256-GCM:+AEAD:+RSA:%COMPAT:%DISABLE_SAFE_RENEGOTIATION:+SIGN-ALL" },
+#endif
 };
 
 #define DTLS_SEND gnutls_record_send
 #define DTLS_RECV gnutls_record_recv
+#define DTLS_FREE gnutls_deinit
 static int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 {
 	gnutls_session_t dtls_ssl;
@@ -391,14 +371,16 @@ static int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 		vpninfo->dtls_attempt_period = 0;
 		return -EINVAL;
 	}
+
 	gnutls_transport_set_ptr(dtls_ssl,
-				 (gnutls_transport_ptr_t)(long) dtls_fd);
+				 (gnutls_transport_ptr_t)(intptr_t)dtls_fd);
+
 	gnutls_record_disable_padding(dtls_ssl);
 	master_secret.data = vpninfo->dtls_secret;
 	master_secret.size = sizeof(vpninfo->dtls_secret);
 	session_id.data = vpninfo->dtls_session_id;
 	session_id.size = sizeof(vpninfo->dtls_session_id);
-	err = gnutls_session_set_premaster(dtls_ssl, GNUTLS_CLIENT, GNUTLS_DTLS0_9,
+	err = gnutls_session_set_premaster(dtls_ssl, GNUTLS_CLIENT, gnutls_dtls_ciphers[cipher].version,
 					   GNUTLS_KX_RSA, gnutls_dtls_ciphers[cipher].cipher,
 					   gnutls_dtls_ciphers[cipher].mac, GNUTLS_COMP_NULL,
 					   &master_secret, &session_id);
@@ -411,19 +393,19 @@ static int start_dtls_handshake(struct openconnect_info *vpninfo, int dtls_fd)
 		return -EINVAL;
 	}
 
-	vpninfo->new_dtls_ssl = dtls_ssl;
+	vpninfo->dtls_ssl = dtls_ssl;
 	return 0;
 }
 
 int dtls_try_handshake(struct openconnect_info *vpninfo)
 {
-	int err = gnutls_handshake(vpninfo->new_dtls_ssl);
+	int err = gnutls_handshake(vpninfo->dtls_ssl);
 
 	if (!err) {
 #ifdef HAVE_GNUTLS_DTLS_SET_DATA_MTU
 		/* Make sure GnuTLS's idea of the MTU is sufficient to take
 		   a full VPN MTU (with 1-byte header) in a data record. */
-		err = gnutls_dtls_set_data_mtu(vpninfo->new_dtls_ssl, vpninfo->actual_mtu + 1);
+		err = gnutls_dtls_set_data_mtu(vpninfo->dtls_ssl, vpninfo->ip_info.mtu + 1);
 		if (err) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Failed to set DTLS MTU: %s\n"),
@@ -435,77 +417,60 @@ int dtls_try_handshake(struct openconnect_info *vpninfo)
 		   we leave enough headroom by adding the worst-case overhead.
 		   We only support AES128-CBC and DES-CBC3-SHA anyway, so
 		   working out the worst case isn't hard. */
-		gnutls_dtls_set_mtu(vpninfo->new_dtls_ssl,
-				    vpninfo->actual_mtu + 1 /* packet + header */
+		gnutls_dtls_set_mtu(vpninfo->dtls_ssl,
+				    vpninfo->ip_info.mtu + 1 /* packet + header */
 				    + 13 /* DTLS header */
 				    + 20 /* biggest supported MAC (SHA1) */
 				    + 16 /* biggest supported IV (AES-128) */
 				    + 16 /* max padding */);
 #endif
 
-		vpn_progress(vpninfo, PRG_INFO, _("Established DTLS connection (using GnuTLS)\n"));
+		vpninfo->dtls_state = DTLS_CONNECTED;
+		vpn_progress(vpninfo, PRG_INFO,
+			     _("Established DTLS connection (using GnuTLS). Ciphersuite %s.\n"),
+			     vpninfo->dtls_cipher);
 
-		if (vpninfo->dtls_ssl) {
-			/* We are replacing an old connection */
-			gnutls_deinit(vpninfo->dtls_ssl);
-			close(vpninfo->dtls_fd);
-			FD_CLR(vpninfo->dtls_fd, &vpninfo->select_rfds);
-			FD_CLR(vpninfo->dtls_fd, &vpninfo->select_wfds);
-			FD_CLR(vpninfo->dtls_fd, &vpninfo->select_efds);
-		}
-		vpninfo->dtls_ssl = vpninfo->new_dtls_ssl;
-		vpninfo->dtls_fd = vpninfo->new_dtls_fd;
-
-		vpninfo->new_dtls_ssl = NULL;
-		vpninfo->new_dtls_fd = -1;
-
-		vpninfo->dtls_times.last_rx = vpninfo->dtls_times.last_tx = time(NULL);
+		vpninfo->dtls_times.last_rekey = vpninfo->dtls_times.last_rx = 
+			vpninfo->dtls_times.last_tx = time(NULL);
 
 		/* XXX: For OpenSSL we explicitly prevent retransmits here. */
 		return 0;
 	}
 
 	if (err == GNUTLS_E_AGAIN) {
-		if (time(NULL) < vpninfo->new_dtls_started + 5)
+		if (time(NULL) < vpninfo->new_dtls_started + 12)
 			return 0;
-		vpn_progress(vpninfo, PRG_TRACE, _("DTLS handshake timed out\n"));
+		vpn_progress(vpninfo, PRG_DEBUG, _("DTLS handshake timed out\n"));
 	}
 
 	vpn_progress(vpninfo, PRG_ERR, _("DTLS handshake failed: %s\n"),
 		     gnutls_strerror(err));
 
-	goto error;
  error:
-	/* Kill the new (failed) connection... */
-	gnutls_deinit(vpninfo->new_dtls_ssl);
-	FD_CLR(vpninfo->new_dtls_fd, &vpninfo->select_rfds);
-	FD_CLR(vpninfo->new_dtls_fd, &vpninfo->select_efds);
-	close(vpninfo->new_dtls_fd);
-	vpninfo->new_dtls_ssl = NULL;
-	vpninfo->new_dtls_fd = -1;
-
-	/* ... and kill the old one too. The only time there'll be a valid
-	   existing session is when it was a rekey, and in that case it's
-	   time for the old one to die. */
-	if (vpninfo->dtls_ssl) {
-		gnutls_deinit(vpninfo->dtls_ssl);
-		close(vpninfo->dtls_fd);
-		FD_CLR(vpninfo->dtls_fd, &vpninfo->select_rfds);
-		FD_CLR(vpninfo->dtls_fd, &vpninfo->select_wfds);
-		FD_CLR(vpninfo->dtls_fd, &vpninfo->select_efds);
-		vpninfo->dtls_ssl = NULL;
-		vpninfo->dtls_fd = -1;
-	}
+	dtls_close(vpninfo);
 
+	vpninfo->dtls_state = DTLS_SLEEPING;
 	time(&vpninfo->new_dtls_started);
 	return -EINVAL;
 }
+
+void dtls_shutdown(struct openconnect_info *vpninfo)
+{
+	dtls_close(vpninfo);
+}
 #endif
 
 int connect_dtls_socket(struct openconnect_info *vpninfo)
 {
 	int dtls_fd, ret, sndbuf;
 
+	/* Sanity check for the removal of new_dtls_{fd,ssl} */
+	if (vpninfo->dtls_fd != -1) {
+		vpn_progress(vpninfo, PRG_ERR, _("DTLS connection attempted with an existing fd\n"));
+		vpninfo->dtls_attempt_period = 0;
+		return -EINVAL;
+	}
+
 	if (!vpninfo->dtls_addr) {
 		vpn_progress(vpninfo, PRG_ERR, _("No DTLS address\n"));
 		vpninfo->dtls_attempt_period = 0;
@@ -531,9 +496,11 @@ int connect_dtls_socket(struct openconnect_info *vpninfo)
 		perror(_("Open UDP socket for DTLS:"));
 		return -EINVAL;
 	}
+	if (vpninfo->protect_socket)
+		vpninfo->protect_socket(vpninfo->cbdata, dtls_fd);
 
-	sndbuf = vpninfo->actual_mtu * 2;
-	setsockopt(dtls_fd, SOL_SOCKET, SO_SNDBUF, &sndbuf, sizeof(sndbuf));
+	sndbuf = vpninfo->ip_info.mtu * 2;
+	setsockopt(dtls_fd, SOL_SOCKET, SO_SNDBUF, (void *)&sndbuf, sizeof(sndbuf));
 
 	if (vpninfo->dtls_local_port) {
 		union {
@@ -560,69 +527,73 @@ int connect_dtls_socket(struct openconnect_info *vpninfo)
 				     _("Unknown protocol family %d. Cannot do DTLS\n"),
 				     vpninfo->peer_addr->sa_family);
 			vpninfo->dtls_attempt_period = 0;
-			close(dtls_fd);
+			closesocket(dtls_fd);
 			return -EINVAL;
 		}
 
 		if (bind(dtls_fd, (struct sockaddr *)&dtls_bind_addr, dtls_bind_addrlen)) {
 			perror(_("Bind UDP socket for DTLS"));
-			close(dtls_fd);
+			closesocket(dtls_fd);
 			return -EINVAL;
 		}
 	}
 
 	if (connect(dtls_fd, vpninfo->dtls_addr, vpninfo->peer_addrlen)) {
 		perror(_("UDP (DTLS) connect:\n"));
-		close(dtls_fd);
+		closesocket(dtls_fd);
 		return -EINVAL;
 	}
 
-	fcntl(dtls_fd, F_SETFD, FD_CLOEXEC);
-	fcntl(dtls_fd, F_SETFL, fcntl(dtls_fd, F_GETFL) | O_NONBLOCK);
+	set_fd_cloexec(dtls_fd);
+	set_sock_nonblock(dtls_fd);
 
 	ret = start_dtls_handshake(vpninfo, dtls_fd);
 	if (ret) {
-		close(dtls_fd);
+		closesocket(dtls_fd);
 		return ret;
 	}
 
-	vpninfo->new_dtls_fd = dtls_fd;
-	if (vpninfo->select_nfds <= dtls_fd)
-		vpninfo->select_nfds = dtls_fd + 1;
+	vpninfo->dtls_state = DTLS_CONNECTING;
 
-	FD_SET(dtls_fd, &vpninfo->select_rfds);
-	FD_SET(dtls_fd, &vpninfo->select_efds);
+	vpninfo->dtls_fd = dtls_fd;
+	monitor_fd_new(vpninfo, dtls);
+	monitor_read_fd(vpninfo, dtls);
+	monitor_except_fd(vpninfo, dtls);
 
 	time(&vpninfo->new_dtls_started);
 
 	return dtls_try_handshake(vpninfo);
 }
 
-static int dtls_restart(struct openconnect_info *vpninfo)
+void dtls_close(struct openconnect_info *vpninfo)
 {
 	if (vpninfo->dtls_ssl) {
-#if defined(DTLS_OPENSSL)
-		SSL_free(vpninfo->dtls_ssl);
-#elif defined(DTLS_GNUTLS)
-		gnutls_deinit(vpninfo->dtls_ssl);
-#endif
-		close(vpninfo->dtls_fd);
-		FD_CLR(vpninfo->dtls_fd, &vpninfo->select_rfds);
-		FD_CLR(vpninfo->dtls_fd, &vpninfo->select_wfds);
-		FD_CLR(vpninfo->dtls_fd, &vpninfo->select_efds);
+		DTLS_FREE(vpninfo->dtls_ssl);
+		closesocket(vpninfo->dtls_fd);
+		unmonitor_read_fd(vpninfo, dtls);
+		unmonitor_write_fd(vpninfo, dtls);
+		unmonitor_except_fd(vpninfo, dtls);
 		vpninfo->dtls_ssl = NULL;
 		vpninfo->dtls_fd = -1;
 	}
+}
 
+int dtls_reconnect(struct openconnect_info *vpninfo)
+{
+	dtls_close(vpninfo);
+	vpninfo->dtls_state = DTLS_SLEEPING;
 	return connect_dtls_socket(vpninfo);
 }
 
-
-int setup_dtls(struct openconnect_info *vpninfo)
+int openconnect_setup_dtls(struct openconnect_info *vpninfo, int dtls_attempt_period)
 {
-	struct vpn_option *dtls_opt = vpninfo->dtls_options;
+	struct oc_vpn_option *dtls_opt = vpninfo->dtls_options;
 	int dtls_port = 0;
 
+	vpninfo->dtls_attempt_period = dtls_attempt_period;
+	if (!dtls_attempt_period)
+		return 0;
+
 #if defined(OPENCONNECT_GNUTLS) && defined(DTLS_OPENSSL)
 	/* If we're using GnuTLS for authentication but OpenSSL for DTLS,
 	   we'll need to initialise OpenSSL now... */
@@ -633,7 +604,7 @@ int setup_dtls(struct openconnect_info *vpninfo)
 #endif
 
 	while (dtls_opt) {
-		vpn_progress(vpninfo, PRG_TRACE,
+		vpn_progress(vpninfo, PRG_DEBUG,
 			     _("DTLS option %s : %s\n"),
 			     dtls_opt->option, dtls_opt->value);
 
@@ -645,6 +616,13 @@ int setup_dtls(struct openconnect_info *vpninfo)
 			int j = atol(dtls_opt->value);
 			if (j && (!vpninfo->dtls_times.dpd || j < vpninfo->dtls_times.dpd))
 				vpninfo->dtls_times.dpd = j;
+		} else if (!strcmp(dtls_opt->option + 7, "Rekey-Method")) {
+			if (!strcmp(dtls_opt->value, "new-tunnel"))
+				vpninfo->dtls_times.rekey_method = REKEY_TUNNEL;
+			else if (!strcmp(dtls_opt->value, "ssl"))
+				vpninfo->dtls_times.rekey_method = REKEY_SSL;
+			else
+				vpninfo->dtls_times.rekey_method = REKEY_NONE;
 		} else if (!strcmp(dtls_opt->option + 7, "Rekey-Time")) {
 			vpninfo->dtls_times.rekey = atol(dtls_opt->value);
 		} else if (!strcmp(dtls_opt->option + 7, "CipherSuite")) {
@@ -657,6 +635,8 @@ int setup_dtls(struct openconnect_info *vpninfo)
 		vpninfo->dtls_attempt_period = 0;
 		return -EINVAL;
 	}
+	if (vpninfo->dtls_times.rekey <= 0)
+		vpninfo->dtls_times.rekey_method = REKEY_NONE;
 
 	vpninfo->dtls_addr = malloc(vpninfo->peer_addrlen);
 	if (!vpninfo->dtls_addr) {
@@ -682,7 +662,7 @@ int setup_dtls(struct openconnect_info *vpninfo)
 	if (connect_dtls_socket(vpninfo))
 		return -EINVAL;
 
-	vpn_progress(vpninfo, PRG_TRACE,
+	vpn_progress(vpninfo, PRG_DEBUG,
 		     _("DTLS initialised. DPD %d, Keepalive %d\n"),
 		     vpninfo->dtls_times.dpd, vpninfo->dtls_times.keepalive);
 
@@ -697,8 +677,25 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	int work_done = 0;
 	char magic_pkt;
 
+	if (vpninfo->dtls_state == DTLS_CONNECTING) {
+		dtls_try_handshake(vpninfo);
+		return 0;
+	}
+
+	if (vpninfo->dtls_state == DTLS_SLEEPING) {
+		int when = vpninfo->new_dtls_started + vpninfo->dtls_attempt_period - time(NULL);
+
+		if (when <= 0) {
+			vpn_progress(vpninfo, PRG_DEBUG, _("Attempt new DTLS connection\n"));
+			connect_dtls_socket(vpninfo);
+		} else if ((when * 1000) < *timeout) {
+			*timeout = when * 1000;
+		}
+		return 0;
+	}
+
 	while (1) {
-		int len = vpninfo->actual_mtu;
+		int len = vpninfo->ip_info.mtu;
 		unsigned char *buf;
 
 		if (!dtls_pkt || len > dtls_pkt_max) {
@@ -730,7 +727,7 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 			break;
 
 		case AC_PKT_DPD_OUT:
-			vpn_progress(vpninfo, PRG_TRACE, _("Got DTLS DPD request\n"));
+			vpn_progress(vpninfo, PRG_DEBUG, _("Got DTLS DPD request\n"));
 
 			/* FIXME: What if the packet doesn't get through? */
 			magic_pkt = AC_PKT_DPD_RESP;
@@ -740,11 +737,11 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 			continue;
 
 		case AC_PKT_DPD_RESP:
-			vpn_progress(vpninfo, PRG_TRACE, _("Got DTLS DPD response\n"));
+			vpn_progress(vpninfo, PRG_DEBUG, _("Got DTLS DPD response\n"));
 			break;
 
 		case AC_PKT_KEEPALIVE:
-			vpn_progress(vpninfo, PRG_TRACE, _("Got DTLS Keepalive\n"));
+			vpn_progress(vpninfo, PRG_DEBUG, _("Got DTLS Keepalive\n"));
 			break;
 
 		default:
@@ -767,29 +764,32 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	}
 
 	switch (keepalive_action(&vpninfo->dtls_times, timeout)) {
-	case KA_REKEY:
+	case KA_REKEY: {
+		int ret;
+
 		vpn_progress(vpninfo, PRG_INFO, _("DTLS rekey due\n"));
 
-		/* There ought to be a method of rekeying DTLS without tearing down
-		   the CSTP session and restarting, but we don't (yet) know it */
-		if (cstp_reconnect(vpninfo)) {
-			vpn_progress(vpninfo, PRG_ERR, _("Reconnect failed\n"));
-			vpninfo->quit_reason = "CSTP reconnect failed";
-			return 1;
+		if (vpninfo->dtls_times.rekey_method == REKEY_SSL) {
+			time(&vpninfo->new_dtls_started);
+			vpninfo->dtls_state = DTLS_CONNECTING;
+			ret = dtls_try_handshake(vpninfo);
+			if (ret) {
+				vpn_progress(vpninfo, PRG_ERR, _("DTLS Rehandshake failed; reconnecting.\n"));
+				return connect_dtls_socket(vpninfo);
+			}
 		}
 
-		if (dtls_restart(vpninfo))
-			vpn_progress(vpninfo, PRG_ERR, _("DTLS rekey failed\n"));
 		return 1;
+	}
 
 	case KA_DPD_DEAD:
 		vpn_progress(vpninfo, PRG_ERR, _("DTLS Dead Peer Detection detected dead peer!\n"));
 		/* Fall back to SSL, and start a new DTLS connection */
-		dtls_restart(vpninfo);
+		dtls_reconnect(vpninfo);
 		return 1;
 
 	case KA_DPD:
-		vpn_progress(vpninfo, PRG_TRACE, _("Send DTLS DPD\n"));
+		vpn_progress(vpninfo, PRG_DEBUG, _("Send DTLS DPD\n"));
 
 		magic_pkt = AC_PKT_DPD_OUT;
 		if (DTLS_SEND(vpninfo->dtls_ssl, &magic_pkt, 1) != 1)
@@ -807,7 +807,7 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 		if (vpninfo->outgoing_queue)
 			break;
 
-		vpn_progress(vpninfo, PRG_TRACE, _("Send DTLS Keepalive\n"));
+		vpn_progress(vpninfo, PRG_DEBUG, _("Send DTLS Keepalive\n"));
 
 		magic_pkt = AC_PKT_KEEPALIVE;
 		if (DTLS_SEND(vpninfo->dtls_ssl, &magic_pkt, 1) != 1)
@@ -822,7 +822,7 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 	}
 
 	/* Service outgoing packet queue */
-	FD_CLR(vpninfo->dtls_fd, &vpninfo->select_wfds);
+	unmonitor_write_fd(vpninfo, dtls);
 	while (vpninfo->outgoing_queue) {
 		struct pkt *this = vpninfo->outgoing_queue;
 		int ret;
@@ -839,7 +839,7 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 			ret = SSL_get_error(vpninfo->dtls_ssl, ret);
 
 			if (ret == SSL_ERROR_WANT_WRITE) {
-				FD_SET(vpninfo->dtls_fd, &vpninfo->select_wfds);
+				monitor_write_fd(vpninfo, dtls);
 				vpninfo->outgoing_queue = this;
 				vpninfo->outgoing_qlen++;
 
@@ -850,7 +850,7 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 					     _("DTLS got write error %d. Falling back to SSL\n"),
 					     ret);
 				openconnect_report_ssl_errors(vpninfo);
-				dtls_restart(vpninfo);
+				dtls_reconnect(vpninfo);
 				vpninfo->outgoing_queue = this;
 				vpninfo->outgoing_qlen++;
 				work_done = 1;
@@ -864,12 +864,12 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 				vpn_progress(vpninfo, PRG_ERR,
 					     _("DTLS got write error: %s. Falling back to SSL\n"),
 					     gnutls_strerror(ret));
-				dtls_restart(vpninfo);
+				dtls_reconnect(vpninfo);
 				vpninfo->outgoing_queue = this;
 				vpninfo->outgoing_qlen++;
 				work_done = 1;
 			} else if (gnutls_record_get_direction(vpninfo->dtls_ssl)) {
-				FD_SET(vpninfo->dtls_fd, &vpninfo->select_wfds);
+				monitor_write_fd(vpninfo, dtls);
 				vpninfo->outgoing_queue = this;
 				vpninfo->outgoing_qlen++;
 			}
@@ -888,10 +888,23 @@ int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout)
 }
 #else /* !HAVE_DTLS */
 #warning Your SSL library does not seem to support Cisco DTLS compatibility
-int setup_dtls(struct openconnect_info *vpninfo)
+int openconnect_setup_dtls(struct openconnect_info *vpninfo, int dtls_attempt_period)
 {
 	vpn_progress(vpninfo, PRG_ERR,
 		     _("Built against SSL library with no Cisco DTLS support\n"));
 	return -EINVAL;
 }
+
+int dtls_reconnect(struct openconnect_info *vpninfo)
+{
+	return -EINVAL;
+}
+
+void dtls_close(struct openconnect_info *vpninfo)
+{
+}
+
+void dtls_shutdown(struct openconnect_info *vpninfo)
+{
+}
 #endif
