@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2012 Intel Corporation.
+ * Copyright © 2008-2014 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -13,26 +13,16 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to:
- *
- *   Free Software Foundation, Inc.
- *   51 Franklin Street, Fifth Floor,
- *   Boston, MA 02110-1301 USA
  */
 
+#include <config.h>
+
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <sys/socket.h>
-#include <netdb.h>
-#include <unistd.h>
 #include <fcntl.h>
 #include <string.h>
 #include <ctype.h>
 #include <stdio.h>
-#include <netinet/in.h>
-#include <arpa/inet.h>
 #include <errno.h>
 #include <stdarg.h>
 #include <stdlib.h>
@@ -67,10 +57,8 @@ static P11KitPin *pin_callback(const char *pin_source, P11KitUri *pin_uri,
 #include "gnutls.h"
 #include "openconnect-internal.h"
 
-/* Helper functions for reading/writing lines over SSL.
-   We could use cURL for the HTTP stuff, but it's overkill */
-
-int openconnect_SSL_write(struct openconnect_info *vpninfo, char *buf, size_t len)
+/* Helper functions for reading/writing lines over SSL. */
+static int openconnect_gnutls_write(struct openconnect_info *vpninfo, char *buf, size_t len)
 {
 	size_t orig_len = len;
 
@@ -78,11 +66,8 @@ int openconnect_SSL_write(struct openconnect_info *vpninfo, char *buf, size_t le
 		int done = gnutls_record_send(vpninfo->https_sess, buf, len);
 		if (done > 0)
 			len -= done;
-		else if (done != GNUTLS_E_AGAIN) {
-			vpn_progress(vpninfo, PRG_ERR, _("Failed to write to SSL socket: %s\n"),
-				     gnutls_strerror(done));
-			return -EIO;
-		} else {
+		else if (done == GNUTLS_E_AGAIN) {
+			/* Wait for something to happen on the socket, or on cmd_fd */
 			fd_set wr_set, rd_set;
 			int maxfd = vpninfo->ssl_fd;
 
@@ -94,35 +79,31 @@ int openconnect_SSL_write(struct openconnect_info *vpninfo, char *buf, size_t le
 			else
 				FD_SET(vpninfo->ssl_fd, &rd_set);
 
-			if (vpninfo->cancel_fd != -1) {
-				FD_SET(vpninfo->cancel_fd, &rd_set);
-				if (vpninfo->cancel_fd > vpninfo->ssl_fd)
-					maxfd = vpninfo->cancel_fd;
-			}
+			cmd_fd_set(vpninfo, &rd_set, &maxfd);
 			select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
-			if (vpninfo->cancel_fd != -1 &&
-			    FD_ISSET(vpninfo->cancel_fd, &rd_set)) {
+			if (is_cancel_pending(vpninfo, &rd_set)) {
 				vpn_progress(vpninfo, PRG_ERR, _("SSL write cancelled\n"));
 				return -EINTR;
 			}
+		} else {
+			vpn_progress(vpninfo, PRG_ERR, _("Failed to write to SSL socket: %s\n"),
+				     gnutls_strerror(done));
+			return -EIO;
 		}
 	}
 	return orig_len;
 }
 
-int openconnect_SSL_read(struct openconnect_info *vpninfo, char *buf, size_t len)
+static int openconnect_gnutls_read(struct openconnect_info *vpninfo, char *buf, size_t len)
 {
 	int done;
 
 	while ((done = gnutls_record_recv(vpninfo->https_sess, buf, len)) < 0) {
-		fd_set wr_set, rd_set;
-		int maxfd = vpninfo->ssl_fd;
+		if (done == GNUTLS_E_AGAIN) {
+			/* Wait for something to happen on the socket, or on cmd_fd */
+			fd_set wr_set, rd_set;
+			int maxfd = vpninfo->ssl_fd;
 
-		if (done != GNUTLS_E_AGAIN) {
-			vpn_progress(vpninfo, PRG_ERR, _("Failed to read from SSL socket: %s\n"),
-				     gnutls_strerror(done));
-			return -EIO;
-		} else {
 			FD_ZERO(&wr_set);
 			FD_ZERO(&rd_set);
 
@@ -131,23 +112,31 @@ int openconnect_SSL_read(struct openconnect_info *vpninfo, char *buf, size_t len
 			else
 				FD_SET(vpninfo->ssl_fd, &rd_set);
 
-			if (vpninfo->cancel_fd != -1) {
-				FD_SET(vpninfo->cancel_fd, &rd_set);
-				if (vpninfo->cancel_fd > vpninfo->ssl_fd)
-					maxfd = vpninfo->cancel_fd;
-			}
+			cmd_fd_set(vpninfo, &rd_set, &maxfd);
 			select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
-			if (vpninfo->cancel_fd != -1 &&
-			    FD_ISSET(vpninfo->cancel_fd, &rd_set)) {
+			if (is_cancel_pending(vpninfo, &rd_set)) {
 				vpn_progress(vpninfo, PRG_ERR, _("SSL read cancelled\n"));
 				return -EINTR;
 			}
+#ifdef GNUTLS_E_PREMATURE_TERMINATION
+		} else if (done == GNUTLS_E_PREMATURE_TERMINATION) {
+			/* We've seen this with HTTP 1.0 responses followed by abrupt
+			   socket closure and no clean SSL shutdown.
+			   https://bugs.launchpad.net/bugs/1225276 */
+			vpn_progress(vpninfo, PRG_DEBUG, _("SSL socket closed uncleanly\n"));
+			return 0;
+#endif
+		} else {
+			vpn_progress(vpninfo, PRG_ERR, _("Failed to read from SSL socket: %s\n"),
+				     gnutls_strerror(done));
+			return -EIO;
 		}
+
 	}
 	return done;
 }
 
-int openconnect_SSL_gets(struct openconnect_info *vpninfo, char *buf, size_t len)
+static int openconnect_gnutls_gets(struct openconnect_info *vpninfo, char *buf, size_t len)
 {
 	int i = 0;
 	int ret;
@@ -172,12 +161,8 @@ int openconnect_SSL_gets(struct openconnect_info *vpninfo, char *buf, size_t len
 				buf[i] = 0;
 				return i;
 			}
-		} else if (ret != GNUTLS_E_AGAIN) {
-			vpn_progress(vpninfo, PRG_ERR, _("Failed to read from SSL socket: %s\n"),
-				     gnutls_strerror(ret));
-			ret = -EIO;
-			break;
-		} else {
+		} else if (ret == GNUTLS_E_AGAIN) {
+			/* Wait for something to happen on the socket, or on cmd_fd */
 			fd_set rd_set, wr_set;
 			int maxfd = vpninfo->ssl_fd;
 
@@ -189,18 +174,18 @@ int openconnect_SSL_gets(struct openconnect_info *vpninfo, char *buf, size_t len
 			else
 				FD_SET(vpninfo->ssl_fd, &rd_set);
 
-			if (vpninfo->cancel_fd != -1) {
-				FD_SET(vpninfo->cancel_fd, &rd_set);
-				if (vpninfo->cancel_fd > vpninfo->ssl_fd)
-					maxfd = vpninfo->cancel_fd;
-			}
+			cmd_fd_set(vpninfo, &rd_set, &maxfd);
 			select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
-			if (vpninfo->cancel_fd != -1 &&
-			    FD_ISSET(vpninfo->cancel_fd, &rd_set)) {
+			if (is_cancel_pending(vpninfo, &rd_set)) {
 				vpn_progress(vpninfo, PRG_ERR, _("SSL read cancelled\n"));
 				ret = -EINTR;
 				break;
 			}
+		} else {
+			vpn_progress(vpninfo, PRG_ERR, _("Failed to read from SSL socket: %s\n"),
+				     gnutls_strerror(ret));
+			ret = -EIO;
+			break;
 		}
 	}
 	buf[i] = 0;
@@ -225,12 +210,27 @@ static int check_certificate_expiry(struct openconnect_info *vpninfo, gnutls_x50
 		reason = _("Client certificate expires soon at");
 
 	if (reason) {
-		struct tm tm;
 		char buf[80];
+#ifdef _WIN32
+		/*
+		 * Windows doesn't have gmtime_r but apparently its gmtime()
+		 * *is* thread-safe because it uses a per-thread static buffer.
+		 * cf. http://sourceforge.net/p/mingw/bugs/1625/
+		 *
+		 * We also explicitly say 'GMT' because %Z would give us the
+		 * Microsoft stupidity "GMT Standard Time". Which is not only
+		 * silly, but also ambiguous because Windows actually says that
+		 * even when it means British Summer Time (GMT+1). And having
+		 * used gmtime() we really *are* giving the time in GMT.
+		 */
+		struct tm *tm = gmtime(&expires);
+		strftime(buf, 80, "%a, %d %b %Y %H:%M:%S GMT", tm);
+#else
+		struct tm tm;
 
 		gmtime_r(&expires, &tm);
 		strftime(buf, 80, "%a, %d %b %Y %T %Z", &tm);
-
+#endif
 		vpn_progress(vpninfo, PRG_ERR, "%s: %s\n", reason, buf);
 	}
 	return 0;
@@ -270,7 +270,7 @@ static int load_datum(struct openconnect_info *vpninfo,
 	}
 #endif
 
-	fd = open(fname, O_RDONLY|O_CLOEXEC);
+	fd = open(fname, O_RDONLY|O_CLOEXEC|O_BINARY);
 	if (fd == -1) {
 		err = errno;
 		vpn_progress(vpninfo, PRG_ERR,
@@ -347,7 +347,15 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo,
 
 	pass = vpninfo->cert_password;
 	while ((err = gnutls_pkcs12_verify_mac(p12, pass)) == GNUTLS_E_MAC_VERIFY_FAILED) {
-		if (pass)
+		if (!pass) {
+			/* OpenSSL's PKCS12_parse() code will try both NULL and "" automatically,
+			 * but GnuTLS requires two separate attempts. */
+			err = gnutls_pkcs12_verify_mac(p12, "");
+			if (!err) {
+				pass = strdup("");
+				break;
+			}
+		} else
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Failed to decrypt PKCS#12 certificate file\n"));
 		free(pass);
@@ -372,7 +380,7 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo,
 		if (pass == vpninfo->cert_password &&
 		    vpninfo->cert_type == CERT_TYPE_UNKNOWN) {
 			/* Make it non-fatal... */
-			level = PRG_TRACE;
+			level = PRG_DEBUG;
 			ret = NOT_PKCS12;
 		}
 
@@ -501,14 +509,12 @@ static int assign_privkey(struct openconnect_info *vpninfo,
 			  unsigned int nr_certs,
 			  uint8_t *free_certs)
 {
-	int i;
-
 	vpninfo->my_certs = gnutls_calloc(nr_certs, sizeof(*certs));
 	if (!vpninfo->my_certs)
 		return GNUTLS_E_MEMORY_ERROR;
 
 	vpninfo->free_my_certs = gnutls_malloc(nr_certs);
-	if (vpninfo->free_my_certs) {
+	if (!vpninfo->free_my_certs) {
 		gnutls_free(vpninfo->my_certs);
 		vpninfo->my_certs = NULL;
 		return GNUTLS_E_MEMORY_ERROR;
@@ -852,6 +858,7 @@ static int import_openssl_pem(struct openconnect_info *vpninfo,
 		if (pass) {
 			vpn_progress(vpninfo, PRG_ERR,  _("Decrypting PEM key failed\n"));
 			free(pass);
+			pass = NULL;
 		}
 		err = request_passphrase(vpninfo, "openconnect_pem",
 					 &pass, _("Enter PEM pass phrase:"));
@@ -957,7 +964,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 
 	/* Load certificate(s) first... */
 	if (cert_is_p11) {
-		vpn_progress(vpninfo, PRG_TRACE,
+		vpn_progress(vpninfo, PRG_DEBUG,
 			     _("Using PKCS#11 certificate %s\n"), cert_url);
 
 		err = gnutls_x509_crt_init(&cert);
@@ -981,7 +988,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 #endif /* HAVE_P11KIT */
 
 	/* OK, not a PKCS#11 certificate so it must be coming from a file... */
-	vpn_progress(vpninfo, PRG_TRACE,
+	vpn_progress(vpninfo, PRG_DEBUG,
 		     _("Using certificate file %s\n"), vpninfo->cert);
 
 	/* Load file contents */
@@ -1056,7 +1063,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 	   them in extra_certs[]. Next we look for the private key ... */
 #if defined(HAVE_P11KIT)
 	if (key_is_p11) {
-		vpn_progress(vpninfo, PRG_TRACE,
+		vpn_progress(vpninfo, PRG_DEBUG,
 			     _("Using PKCS#11 key %s\n"), key_url);
 
 		err = gnutls_pkcs11_privkey_init(&p11key);
@@ -1199,7 +1206,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		gnutls_free(fdata.data);
 		fdata.data = NULL;
 
-		vpn_progress(vpninfo, PRG_TRACE,
+		vpn_progress(vpninfo, PRG_DEBUG,
 			     _("Using private key file %s\n"), vpninfo->sslkey);
 
 		ret = load_datum(vpninfo, &fdata, vpninfo->sslkey);
@@ -1782,11 +1789,58 @@ static int verify_peer(gnutls_session_t session)
 		return GNUTLS_E_CERTIFICATE_ERROR;
 	}
 
-	if (!reason && !gnutls_x509_crt_check_hostname(cert, vpninfo->hostname))
-		reason = _("certificate does not match hostname");
+	if (reason)
+		goto done;
 
+	if (!gnutls_x509_crt_check_hostname(cert, vpninfo->hostname)) {
+		int i, ret;
+		unsigned char addrbuf[sizeof(struct in6_addr)];
+		unsigned char certaddr[sizeof(struct in6_addr)];
+		size_t addrlen = 0, certaddrlen;
+
+		/* gnutls_x509_crt_check_hostname() doesn't cope with IPv6 literals
+		   in URI form with surrounding [] so we must check for ourselves. */
+		if (vpninfo->hostname[0] == '[' &&
+		    vpninfo->hostname[strlen(vpninfo->hostname)-1] == ']') {
+			char *p = &vpninfo->hostname[strlen(vpninfo->hostname)-1];
+			*p = 0;
+			if (inet_pton(AF_INET6, vpninfo->hostname + 1, addrbuf) > 0)
+				addrlen = 16;
+			*p = ']';
+		}
+#if GNUTLS_VERSION_NUMBER < 0x030306
+		/* And before 3.3.6 it didn't check IP addresses at all. */
+		else if (inet_pton(AF_INET, vpninfo->hostname, addrbuf) > 0)
+			addrlen = 4;
+		else if (inet_pton(AF_INET6, vpninfo->hostname, addrbuf) > 0)
+			addrlen = 16;
+#endif
+		if (!addrlen) {
+			/* vpninfo->hostname was not a bare IP address. Nothing to do */
+			goto badhost;
+		}
+
+		for (i = 0; ; i++) {
+			certaddrlen = sizeof(certaddr);
+			ret = gnutls_x509_crt_get_subject_alt_name(cert, i, certaddr,
+								   &certaddrlen, NULL);
+			/* If this happens, it wasn't an IP address. */
+			if (ret == GNUTLS_E_SHORT_MEMORY_BUFFER)
+				continue;
+			if (ret < 0)
+				break;
+			if (ret != GNUTLS_SAN_IPADDRESS)
+				continue;
+			if (certaddrlen == addrlen && !memcmp(addrbuf, certaddr, addrlen))
+				goto done;
+		}
+	badhost:
+		reason = _("certificate does not match hostname");
+	}
+ done:
 	if (reason) {
-		vpn_progress(vpninfo, PRG_INFO, "Server certificate verify failed: %s\n",
+		vpn_progress(vpninfo, PRG_INFO,
+			     _("Server certificate verify failed: %s\n"),
 			     reason);
 		if (vpninfo->validate_peer_cert)
 			err = vpninfo->validate_peer_cert(vpninfo->cbdata,
@@ -1802,10 +1856,29 @@ static int verify_peer(gnutls_session_t session)
 }
 
 
+/* The F5 firewall is confused when the TLS client hello is between
+ * 256 and 512 bytes. By disabling several TLS options we force the
+ * client hello to be < 256 bytes. We don't do that in gnutls versions
+ * >= 3.2.9 as there the %COMPAT keyword ensures that the client hello
+ * will be outside that range.
+ */
+#if GNUTLS_VERSION_NUMBER >= 0x030209
+# define DEFAULT_PRIO "NORMAL:-VERS-SSL3.0:%COMPAT"
+#else
+# define _DEFAULT_PRIO "NORMAL:-VERS-TLS-ALL:+VERS-TLS1.0:" \
+	"%COMPAT:%DISABLE_SAFE_RENEGOTIATION:%LATEST_RECORD_VERSION"
+# if GNUTLS_VERSION_MAJOR >= 3
+#  define DEFAULT_PRIO _DEFAULT_PRIO":-CURVE-ALL"
+#else
+#  define DEFAULT_PRIO _DEFAULT_PRIO
+# endif
+#endif
+
 int openconnect_open_https(struct openconnect_info *vpninfo)
 {
 	int ssl_sock = -1;
 	int err;
+	const char * prio;
 
 	if (vpninfo->https_sess)
 		return 0;
@@ -1851,7 +1924,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 					gnutls_free(datum.data);
 					gnutls_certificate_free_credentials(vpninfo->https_cred);
 					vpninfo->https_cred = NULL;
-					close(ssl_sock);
+					closesocket(ssl_sock);
 					return -ENOMEM;
 				}
 				err = gnutls_x509_crt_list_import(certs, &nr_certs, &datum,
@@ -1872,7 +1945,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 						     gnutls_strerror(err));
 					gnutls_certificate_free_credentials(vpninfo->https_cred);
 					vpninfo->https_cred = NULL;
-					close(ssl_sock);
+					closesocket(ssl_sock);
 					return -EINVAL;
 				}
 			}
@@ -1888,7 +1961,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 					     vpninfo->cafile, gnutls_strerror(err));
 				gnutls_certificate_free_credentials(vpninfo->https_cred);
 				vpninfo->https_cred = NULL;
-				close(ssl_sock);
+				closesocket(ssl_sock);
 				return -EINVAL;
 			}
 		}
@@ -1900,7 +1973,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 					     _("Loading certificate failed. Aborting.\n"));
 				gnutls_certificate_free_credentials(vpninfo->https_cred);
 				vpninfo->https_cred = NULL;
-				close(ssl_sock);
+				closesocket(ssl_sock);
 				return err;
 			}
 		}
@@ -1911,31 +1984,60 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 	if (vpninfo->my_pkey == OPENCONNECT_TPM_PKEY)
 		gnutls_sign_callback_set(vpninfo->https_sess, gtls2_tpm_sign_cb, vpninfo);
 #endif
+	/* We depend on 3.2.9 because that has the workaround for the
+	   obnoxious F5 firewall that drops packets of certain sizes */
+	if (gnutls_check_version("3.2.9") &&
+	    string_is_hostname(vpninfo->hostname))
+		gnutls_server_name_set(vpninfo->https_sess, GNUTLS_NAME_DNS,
+				       vpninfo->hostname,
+				       strlen(vpninfo->hostname));
+
+	if (vpninfo->pfs) {
+		prio = DEFAULT_PRIO":-RSA";
+	} else {
+		prio = DEFAULT_PRIO;
+	}
 
 	err = gnutls_priority_set_direct(vpninfo->https_sess,
-					 "NORMAL:-VERS-TLS-ALL:+VERS-TLS1.0:"
-#if GNUTLS_VERSION_MAJOR >= 3
-					 "-CURVE-ALL:"
-#endif
-					 "%COMPAT:%DISABLE_SAFE_RENEGOTIATION:%LATEST_RECORD_VERSION",
-					 NULL);
+					prio, NULL);
 	if (err) {
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Failed to set TLS priority string: %s\n"),
 			     gnutls_strerror(err));
 		gnutls_deinit(vpninfo->https_sess);
 		vpninfo->https_sess = NULL;
-		close(ssl_sock);
+		closesocket(ssl_sock);
 		return -EIO;
 	}
 
 	gnutls_record_disable_padding(vpninfo->https_sess);
 	gnutls_credentials_set(vpninfo->https_sess, GNUTLS_CRD_CERTIFICATE, vpninfo->https_cred);
-	gnutls_transport_set_ptr(vpninfo->https_sess, /* really? */(gnutls_transport_ptr_t)(long) ssl_sock);
+	gnutls_transport_set_ptr(vpninfo->https_sess,(gnutls_transport_ptr_t)(intptr_t)ssl_sock);
 
 	vpn_progress(vpninfo, PRG_INFO, _("SSL negotiation with %s\n"),
 		     vpninfo->hostname);
 
+
+	err = cstp_handshake(vpninfo, 1);
+	if (err)
+		return err;
+
+	vpninfo->ssl_fd = ssl_sock;
+
+	vpninfo->ssl_read = openconnect_gnutls_read;
+	vpninfo->ssl_write = openconnect_gnutls_write;
+	vpninfo->ssl_gets = openconnect_gnutls_gets;
+
+	return 0;
+}
+
+int cstp_handshake(struct openconnect_info *vpninfo, unsigned init)
+{
+	int err;
+	int ssl_sock = -1;
+
+	ssl_sock = (intptr_t)gnutls_transport_get_ptr(vpninfo->https_sess);
+
 	while ((err = gnutls_handshake(vpninfo->https_sess))) {
 		if (err == GNUTLS_E_AGAIN) {
 			fd_set rd_set, wr_set;
@@ -1949,18 +2051,13 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 			else
 				FD_SET(ssl_sock, &rd_set);
 
-			if (vpninfo->cancel_fd != -1) {
-				FD_SET(vpninfo->cancel_fd, &rd_set);
-				if (vpninfo->cancel_fd > ssl_sock)
-					maxfd = vpninfo->cancel_fd;
-			}
+			cmd_fd_set(vpninfo, &rd_set, &maxfd);
 			select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
-			if (vpninfo->cancel_fd != -1 &&
-			    FD_ISSET(vpninfo->cancel_fd, &rd_set)) {
+			if (is_cancel_pending(vpninfo, &rd_set)) {
 				vpn_progress(vpninfo, PRG_ERR, _("SSL connection cancelled\n"));
 				gnutls_deinit(vpninfo->https_sess);
 				vpninfo->https_sess = NULL;
-				close(ssl_sock);
+				closesocket(ssl_sock);
 				return -EINTR;
 			}
 		} else if (err == GNUTLS_E_INTERRUPTED || gnutls_error_is_fatal(err)) {
@@ -1968,20 +2065,23 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 							 gnutls_strerror(err));
 			gnutls_deinit(vpninfo->https_sess);
 			vpninfo->https_sess = NULL;
-			close(ssl_sock);
+			closesocket(ssl_sock);
 			return -EIO;
 		} else {
 			/* non-fatal error or warning. Ignore it and continue */
-			vpn_progress(vpninfo, PRG_TRACE,
+			vpn_progress(vpninfo, PRG_DEBUG,
 				     _("GnuTLS non-fatal return during handshake: %s\n"),
 				     gnutls_strerror(err));
 		}
 	}
 
-	vpninfo->ssl_fd = ssl_sock;
-
-	vpn_progress(vpninfo, PRG_INFO, _("Connected to HTTPS on %s\n"),
-		     vpninfo->hostname);
+	if (init) {
+		vpn_progress(vpninfo, PRG_INFO, _("Connected to HTTPS on %s\n"),
+			     vpninfo->hostname);
+	} else {
+		vpn_progress(vpninfo, PRG_INFO, _("Renegotiated SSL on %s\n"),
+			     vpninfo->hostname);
+	}
 
 	return 0;
 }
@@ -1997,10 +2097,10 @@ void openconnect_close_https(struct openconnect_info *vpninfo, int final)
 		vpninfo->https_sess = NULL;
 	}
 	if (vpninfo->ssl_fd != -1) {
-		close(vpninfo->ssl_fd);
-		FD_CLR(vpninfo->ssl_fd, &vpninfo->select_rfds);
-		FD_CLR(vpninfo->ssl_fd, &vpninfo->select_wfds);
-		FD_CLR(vpninfo->ssl_fd, &vpninfo->select_efds);
+		closesocket(vpninfo->ssl_fd);
+		unmonitor_read_fd(vpninfo, ssl);
+		unmonitor_write_fd(vpninfo, ssl);
+		unmonitor_except_fd(vpninfo, ssl);
 		vpninfo->ssl_fd = -1;
 	}
 	if (final && vpninfo->https_cred) {
@@ -2069,6 +2169,9 @@ void openconnect_close_https(struct openconnect_info *vpninfo, int final)
 
 void openconnect_init_ssl(void)
 {
+#ifdef _WIN32
+	openconnect__win32_sock_init();
+#endif
 	gnutls_global_init();
 }
 
@@ -2085,6 +2188,19 @@ int openconnect_sha1(unsigned char *result, void *data, int datalen)
 	return 0;
 }
 
+int openconnect_md5(unsigned char *result, void *data, int datalen)
+{
+	gnutls_datum_t d;
+	size_t md5len = MD5_SIZE;
+
+	d.data = data;
+	d.size = datalen;
+	if (gnutls_fingerprint(GNUTLS_DIG_MD5, &d, result, &md5len))
+		return -1;
+
+	return 0;
+}
+
 int openconnect_random(void *bytes, int len)
 {
 	if (gnutls_rnd(GNUTLS_RND_RANDOM, bytes, len))
@@ -2183,7 +2299,7 @@ static P11KitPin *pin_callback(const char *pin_source, P11KitUri *pin_uri,
 	o.label = (char *)_("Enter PIN:");
 	o.value = NULL;
 
-	ret = vpninfo->process_auth_form(vpninfo->cbdata, &f);
+	ret = process_auth_form(vpninfo, &f);
 	if (ret || !o.value)
 		return NULL;
 
