@@ -7,6 +7,8 @@
  * explicit permission.
  */
 
+#include <config.h>
+
 #ifndef HAVE_GNUTLS_PKCS12_SIMPLE_PARSE
 
 #include <string.h>
@@ -33,12 +35,6 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
- * USA
- *
  */
 
 
