@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2012 Intel Corporation.
+ * Copyright © 2008-2014 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -13,13 +13,6 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to:
- *
- *   Free Software Foundation, Inc.
- *   51 Franklin Street, Fifth Floor,
- *   Boston, MA 02110-1301 USA
  */
 
 /*
@@ -27,6 +20,8 @@
  * Carolin Latze <latze@angry-red-pla.net> and Tobias Soder
  */
 
+#include <config.h>
+
 #include <errno.h>
 #include <string.h>
 
@@ -107,7 +102,7 @@ static int tpm_sign_fn(gnutls_privkey_t key, void *_vpninfo,
 	TSS_HHASH hash;
 	int err;
 
-	vpn_progress(vpninfo, PRG_TRACE,
+	vpn_progress(vpninfo, PRG_DEBUG,
 		     _("TPM sign function called for %d bytes.\n"),
 		     data->size);
 
