@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2012 Intel Corporation.
+ * Copyright © 2008-2014 Intel Corporation.
  * Copyright © 2013 John Morrissey <jwm@horde.net>
  *
  * Authors: David Woodhouse <dwmw2@infradead.org>
@@ -14,18 +14,15 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to:
- *
- *   Free Software Foundation, Inc.
- *   51 Franklin Street, Fifth Floor,
- *   Boston, MA 02110-1301 USA
  */
 
+#include <config.h>
+
 #include <string.h>
 #include <errno.h>
 #include <stdlib.h>
+#include <unistd.h>
+#include <fcntl.h>
 
 #ifdef HAVE_LIBSTOKEN
 #include <stoken.h>
@@ -36,6 +33,7 @@
 #endif
 
 #include <libxml/tree.h>
+#include <zlib.h>
 
 #include "openconnect-internal.h"
 
@@ -48,41 +46,63 @@ struct openconnect_info *openconnect_vpninfo_new(char *useragent,
 {
 	struct openconnect_info *vpninfo = calloc(sizeof(*vpninfo), 1);
 
-	vpninfo->ssl_fd = -1;
+	if (!vpninfo)
+		return NULL;
+
+#ifndef _WIN32
+	vpninfo->tun_fd = -1;
+#endif
+	vpninfo->ssl_fd = vpninfo->dtls_fd = -1;
+	vpninfo->cmd_fd = vpninfo->cmd_fd_write = -1;
 	vpninfo->cert_expire_warning = 60 * 86400;
+	vpninfo->deflate = 1;
+	vpninfo->max_qlen = 10;
+	vpninfo->localname = strdup("localhost");
 	vpninfo->useragent = openconnect_create_useragent(useragent);
 	vpninfo->validate_peer_cert = validate_peer_cert;
 	vpninfo->write_new_config = write_new_config;
 	vpninfo->process_auth_form = process_auth_form;
 	vpninfo->progress = progress;
 	vpninfo->cbdata = privdata ? : vpninfo;
-	vpninfo->cancel_fd = -1;
 	vpninfo->xmlpost = 1;
 	openconnect_set_reported_os(vpninfo, NULL);
 
+	if (!vpninfo->localname || !vpninfo->useragent)
+		goto err;
+
 #ifdef ENABLE_NLS
 	bindtextdomain("openconnect", LOCALEDIR);
 #endif
 
 	return vpninfo;
+
+err:
+	free(vpninfo->localname);
+	free(vpninfo->useragent);
+	free(vpninfo);
+	return NULL;
 }
 
 int openconnect_set_reported_os(struct openconnect_info *vpninfo, const char *os)
 {
 	if (!os) {
 #if defined(__APPLE__)
-		os = "mac";
+		os = "mac-intel";
+#elif defined(__ANDROID__)
+		os = "android";
 #else
 		os = sizeof(long) > 4 ? "linux-64" : "linux";
 #endif
 	}
 
-	/* FIXME: is there a special platname for 64-bit Windows? */
-	if (!strcmp(os, "mac"))
+	if (!strcmp(os, "mac-intel"))
 		vpninfo->csd_xmltag = "csdMac";
 	else if (!strcmp(os, "linux") || !strcmp(os, "linux-64"))
 		vpninfo->csd_xmltag = "csdLinux";
-	else if (!strcmp(os, "win"))
+	else if (!strcmp(os, "android") || !strcmp(os, "apple-ios")) {
+		vpninfo->csd_xmltag = "csdLinux";
+		vpninfo->csd_nostub = 1;
+	} else if (!strcmp(os, "win"))
 		vpninfo->csd_xmltag = "csd";
 	else
 		return -EINVAL;
@@ -91,9 +111,19 @@ int openconnect_set_reported_os(struct openconnect_info *vpninfo, const char *os
 	return 0;
 }
 
-static void free_optlist(struct vpn_option *opt)
+void openconnect_set_mobile_info(struct openconnect_info *vpninfo,
+				 char *mobile_platform_version,
+				 char *mobile_device_type,
+				 char *mobile_device_uniqueid)
 {
-	struct vpn_option *next;
+	vpninfo->mobile_platform_version = mobile_platform_version;
+	vpninfo->mobile_device_type = mobile_device_type;
+	vpninfo->mobile_device_uniqueid = mobile_device_uniqueid;
+}
+
+static void free_optlist(struct oc_vpn_option *opt)
+{
+	struct oc_vpn_option *next;
 
 	for (; opt; opt = next) {
 		next = opt->next;
@@ -106,20 +136,46 @@ static void free_optlist(struct vpn_option *opt)
 void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 {
 	openconnect_close_https(vpninfo, 1);
+	dtls_shutdown(vpninfo);
+	if (vpninfo->cmd_fd_write != -1) {
+		close(vpninfo->cmd_fd);
+		close(vpninfo->cmd_fd_write);
+	}
+#ifdef _WIN32
+	if (vpninfo->cmd_event)
+		CloseHandle(vpninfo->cmd_event);
+	if (vpninfo->ssl_event)
+		CloseHandle(vpninfo->ssl_event);
+	if (vpninfo->dtls_event)
+		CloseHandle(vpninfo->dtls_event);
+#endif
 	free(vpninfo->peer_addr);
 	free_optlist(vpninfo->cookies);
 	free_optlist(vpninfo->cstp_options);
 	free_optlist(vpninfo->dtls_options);
+	cstp_free_splits(vpninfo);
 	free(vpninfo->hostname);
 	free(vpninfo->urlpath);
 	free(vpninfo->redirect_url);
+	free(vpninfo->cookie);
 	free(vpninfo->proxy_type);
 	free(vpninfo->proxy);
+	free(vpninfo->proxy_user);
+	free(vpninfo->proxy_pass);
+	free(vpninfo->vpnc_script);
+	free(vpninfo->cafile);
+	free(vpninfo->servercert);
+	free(vpninfo->ifname);
+	free(vpninfo->dtls_cipher);
+	free(vpninfo->dtls_addr);
 
 	if (vpninfo->csd_scriptname) {
 		unlink(vpninfo->csd_scriptname);
 		free(vpninfo->csd_scriptname);
 	}
+	free(vpninfo->mobile_platform_version);
+	free(vpninfo->mobile_device_type);
+	free(vpninfo->mobile_device_uniqueid);
 	free(vpninfo->csd_token);
 	free(vpninfo->csd_ticket);
 	free(vpninfo->csd_stuburl);
@@ -128,11 +184,12 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 	free(vpninfo->csd_preurl);
 	if (vpninfo->opaque_srvdata)
 		xmlFreeNode(vpninfo->opaque_srvdata);
+	free(vpninfo->profile_url);
+	free(vpninfo->profile_sha1);
 
 	/* These are const in openconnect itself, but for consistency of
 	   the library API we do take ownership of the strings we're given,
 	   and thus we have to free them too. */
-	free((void *)vpninfo->cafile);
 	if (vpninfo->cert != vpninfo->sslkey)
 		free((void *)vpninfo->sslkey);
 	free((void *)vpninfo->cert);
@@ -144,7 +201,9 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 #endif
 		vpninfo->peer_cert = NULL;
 	}
+	free(vpninfo->localname);
 	free(vpninfo->useragent);
+	free(vpninfo->authgroup);
 #ifdef HAVE_LIBSTOKEN
 	if (vpninfo->stoken_pin)
 		free(vpninfo->stoken_pin);
@@ -155,7 +214,12 @@ void openconnect_vpninfo_free(struct openconnect_info *vpninfo)
 	if (vpninfo->oath_secret)
 		oath_done();
 #endif
-	/* No need to free deflate streams; they weren't initialised */
+
+	/* These check strm->state so they are safe to call multiple times */
+	inflateEnd(&vpninfo->inflate_strm);
+	deflateEnd(&vpninfo->deflate_strm);
+
+	free(vpninfo->deflate_pkt);
 	free(vpninfo);
 }
 
@@ -197,6 +261,44 @@ void openconnect_set_cafile(struct openconnect_info *vpninfo, char *cafile)
 	vpninfo->cafile = cafile;
 }
 
+void openconnect_set_server_cert_sha1(struct openconnect_info *vpninfo, char *servercert)
+{
+	vpninfo->servercert = servercert;
+}
+
+const char *openconnect_get_ifname(struct openconnect_info *vpninfo)
+{
+	return vpninfo->ifname;
+}
+
+void openconnect_set_reqmtu(struct openconnect_info *vpninfo, int reqmtu)
+{
+	vpninfo->reqmtu = reqmtu;
+}
+
+void openconnect_set_dpd(struct openconnect_info *vpninfo, int min_seconds)
+{
+	/* Make sure (ka->dpd / 2), our computed midway point, isn't 0 */
+	if (!min_seconds || min_seconds >= 2)
+		vpninfo->dtls_times.dpd = vpninfo->ssl_times.dpd = min_seconds;
+	else if (min_seconds == 1)
+		vpninfo->dtls_times.dpd = vpninfo->ssl_times.dpd = 2;
+}
+
+int openconnect_get_ip_info(struct openconnect_info *vpninfo,
+			    const struct oc_ip_info **info,
+			    const struct oc_vpn_option **cstp_options,
+			    const struct oc_vpn_option **dtls_options)
+{
+	if (info)
+		*info = &vpninfo->ip_info;
+	if (cstp_options)
+		*cstp_options = vpninfo->cstp_options;
+	if (dtls_options)
+		*dtls_options = vpninfo->dtls_options;
+	return 0;
+}
+
 void openconnect_setup_csd(struct openconnect_info *vpninfo, uid_t uid, int silent, char *wrapper)
 {
 	vpninfo->uid_csd = uid;
@@ -204,6 +306,11 @@ void openconnect_setup_csd(struct openconnect_info *vpninfo, uid_t uid, int sile
 	vpninfo->csd_wrapper = wrapper;
 }
 
+void openconnect_set_xmlpost(struct openconnect_info *vpninfo, int enable)
+{
+	vpninfo->xmlpost = enable;
+}
+
 void openconnect_set_client_cert(struct openconnect_info *vpninfo, char *cert, char *sslkey)
 {
 	vpninfo->cert = cert;
@@ -236,6 +343,7 @@ void openconnect_clear_cookie(struct openconnect_info *vpninfo)
 
 void openconnect_reset_ssl(struct openconnect_info *vpninfo)
 {
+	vpninfo->got_cancel_cmd = 0;
 	openconnect_close_https(vpninfo, 0);
 	if (vpninfo->peer_addr) {
 		free(vpninfo->peer_addr);
@@ -276,9 +384,36 @@ void openconnect_set_cert_expiry_warning(struct openconnect_info *vpninfo,
 	vpninfo->cert_expire_warning = seconds;
 }
 
+void openconnect_set_pfs(struct openconnect_info *vpninfo,
+					  	unsigned val)
+{
+	vpninfo->pfs = val;
+}
+
 void openconnect_set_cancel_fd(struct openconnect_info *vpninfo, int fd)
 {
-	vpninfo->cancel_fd = fd;
+	vpninfo->cmd_fd = fd;
+}
+
+int openconnect_setup_cmd_pipe(struct openconnect_info *vpninfo)
+{
+	int pipefd[2];
+
+#ifdef _WIN32
+	if (dumb_socketpair(pipefd, 0))
+		return -EIO;
+#else
+	if (pipe(pipefd) < 0)
+		return -EIO;
+#endif
+	if (set_sock_nonblock(pipefd[0]) || set_sock_nonblock(pipefd[1])) {
+		close(pipefd[0]);
+		close(pipefd[1]);
+		return -EIO;
+	}
+	vpninfo->cmd_fd = pipefd[0];
+	vpninfo->cmd_fd_write = pipefd[1];
+	return vpninfo->cmd_fd_write;
 }
 
 const char *openconnect_get_version(void)
@@ -328,7 +463,7 @@ int openconnect_has_stoken_support(void)
 int openconnect_has_oath_support(void)
 {
 #ifdef HAVE_LIBOATH
-	return 1;
+	return 2;
 #else
 	return 0;
 #endif
@@ -359,7 +494,7 @@ static int set_libstoken_mode(struct openconnect_info *vpninfo,
 #endif
 }
 
-static int set_oath_mode(struct openconnect_info *vpninfo,
+static int set_totp_mode(struct openconnect_info *vpninfo,
 			 const char *token_str)
 {
 #ifdef HAVE_LIBOATH
@@ -388,6 +523,48 @@ static int set_oath_mode(struct openconnect_info *vpninfo,
 #endif
 }
 
+static int set_hotp_mode(struct openconnect_info *vpninfo,
+			 const char *token_str)
+{
+#ifdef HAVE_LIBOATH
+	int ret, toklen;
+	char *p;
+
+	ret = oath_init();
+	if (ret != OATH_OK)
+		return -EIO;
+
+	toklen = strlen(token_str);
+	p = strrchr(token_str, ',');
+	if (p) {
+		long counter;
+		toklen = p - token_str;
+		p++;
+		counter = strtol(p, &p, 0);
+		if (counter < 0 || *p)
+			return -EINVAL;
+		vpninfo->token_time = counter;
+	}
+
+	if (strncasecmp(token_str, "base32:", strlen("base32:")) == 0) {
+		ret = oath_base32_decode(token_str + strlen("base32:"),
+					 toklen - strlen("base32:"),
+					 &vpninfo->oath_secret,
+					 &vpninfo->oath_secret_len);
+		if (ret != OATH_OK)
+			return -EINVAL;
+	} else {
+		vpninfo->oath_secret = strdup(token_str);
+		vpninfo->oath_secret_len = toklen;
+	}
+
+	vpninfo->token_mode = OC_TOKEN_MODE_HOTP;
+	return 0;
+#else
+	return -EOPNOTSUPP;
+#endif
+}
+
 /*
  * Enable software token generation.
  *
@@ -417,7 +594,10 @@ int openconnect_set_token_mode(struct openconnect_info *vpninfo,
 		return set_libstoken_mode(vpninfo, token_str);
 
 	case OC_TOKEN_MODE_TOTP:
-		return set_oath_mode(vpninfo, token_str);
+		return set_totp_mode(vpninfo, token_str);
+
+	case OC_TOKEN_MODE_HOTP:
+		return set_hotp_mode(vpninfo, token_str);
 
 	default:
 		return -EOPNOTSUPP;
@@ -449,3 +629,36 @@ int openconnect_set_stoken_mode(struct openconnect_info *vpninfo,
 
 	return openconnect_set_token_mode(vpninfo, token_mode, token_str);
 }
+
+void openconnect_set_protect_socket_handler(struct openconnect_info *vpninfo,
+					    openconnect_protect_socket_vfn protect_socket)
+{
+	vpninfo->protect_socket = protect_socket;
+}
+
+void openconnect_set_stats_handler(struct openconnect_info *vpninfo,
+				   openconnect_stats_vfn stats_handler)
+{
+	vpninfo->stats_handler = stats_handler;
+}
+
+/* Set up a traditional OS-based tunnel device, optionally specified in 'ifname'. */
+int openconnect_setup_tun_device(struct openconnect_info *vpninfo, char *vpnc_script, char *ifname)
+{
+	intptr_t tun_fd;
+
+	vpninfo->vpnc_script = vpnc_script;
+	vpninfo->ifname = ifname;
+
+	set_script_env(vpninfo);
+	script_config_tun(vpninfo, "pre-init");
+
+	tun_fd = os_setup_tun(vpninfo);
+	if (tun_fd < 0)
+		return tun_fd;
+
+	setenv("TUNDEV", vpninfo->ifname, 1);
+	script_config_tun(vpninfo, "connect");
+
+	return openconnect_setup_tun_fd(vpninfo, tun_fd);
+}
