@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2012 Intel Corporation.
+ * Copyright © 2008-2014 Intel Corporation.
  * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
  * Copyright © 2013 John Morrissey <jwm@horde.net>
  *
@@ -15,13 +15,6 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to:
- *
- *   Free Software Foundation, Inc.
- *   51 Franklin Street, Fifth Floor,
- *   Boston, MA 02110-1301 USA
  */
 
 #ifndef __OPENCONNECT_INTERNAL_H__
@@ -29,6 +22,23 @@
 
 #define __OPENCONNECT_PRIVATE__
 
+#ifdef _WIN32
+#include <winsock2.h>
+#include <ws2tcpip.h>
+#ifndef SECURITY_WIN32
+#define SECURITY_WIN32 1
+#endif
+#include <security.h>
+#else
+#include <sys/types.h>
+#include <sys/socket.h>
+#include <sys/select.h>
+#include <netdb.h>
+#include <netinet/in.h>
+#include <arpa/inet.h>
+#include <fcntl.h>
+#endif
+
 #include "openconnect.h"
 
 #if defined(OPENCONNECT_OPENSSL) || defined(DTLS_OPENSSL)
@@ -54,8 +64,6 @@
 
 #include <zlib.h>
 #include <stdint.h>
-#include <sys/socket.h>
-#include <sys/select.h>
 #include <sys/time.h>
 #include <sys/types.h>
 #include <unistd.h>
@@ -68,6 +76,10 @@
 #include <stoken.h>
 #endif
 
+#ifdef HAVE_GSSAPI
+#include GSSAPI_HDR
+#endif
+
 #ifdef ENABLE_NLS
 #include <locale.h>
 #include <libintl.h>
@@ -91,11 +103,9 @@ struct pkt {
 	unsigned char data[];
 };
 
-struct vpn_option {
-	char *option;
-	char *value;
-	struct vpn_option *next;
-};
+#define REKEY_NONE      0
+#define REKEY_TUNNEL    1
+#define REKEY_SSL       2
 
 #define KA_NONE		0
 #define KA_DPD		1
@@ -103,27 +113,36 @@ struct vpn_option {
 #define KA_KEEPALIVE	3
 #define KA_REKEY	4
 
+#define DTLS_NOSECRET	0
+#define DTLS_DISABLED	1
+#define DTLS_SLEEPING	2
+#define DTLS_CONNECTING	3
+#define DTLS_CONNECTED	4
+
 struct keepalive_info {
 	int dpd;
 	int keepalive;
 	int rekey;
+	int rekey_method;
 	time_t last_rekey;
 	time_t last_tx;
 	time_t last_rx;
 	time_t last_dpd;
 };
 
-struct split_include {
-	char *route;
-	struct split_include *next;
-};
-
 struct pin_cache {
 	struct pin_cache *next;
 	char *token;
 	char *pin;
 };
 
+struct oc_text_buf {
+	char *data;
+	int pos;
+	int buf_len;
+	int error;
+};
+
 #define RECONNECT_INTERVAL_MIN	10
 #define RECONNECT_INTERVAL_MAX	100
 
@@ -136,12 +155,35 @@ struct pin_cache {
 #define REDIR_TYPE_NEWHOST	1
 #define REDIR_TYPE_LOCAL	2
 
+#define AUTH_TYPE_GSSAPI	0
+#define AUTH_TYPE_NTLM		1
+#define AUTH_TYPE_DIGEST	2
+#define AUTH_TYPE_BASIC		3
+
+#define MAX_AUTH_TYPES		4
+
+#define AUTH_DISABLED		-2
+#define AUTH_FAILED		-1	/* Failed */
+#define AUTH_UNSEEN		0	/* Server has not offered it */
+#define AUTH_AVAILABLE		1	/* Server has offered it, we have not tried it */
+	/* Individual auth types may use 2 onwards for their own state */
+#define AUTH_IN_PROGRESS	2	/* In-progress attempt */
+
+struct proxy_auth_state {
+	int state;
+	char *challenge;
+};
+
 struct openconnect_info {
 	char *redirect_url;
 	int redirect_type;
 
 	const char *csd_xmltag;
+	int csd_nostub;
 	const char *platname;
+	char *mobile_platform_version;
+	char *mobile_device_type;
+	char *mobile_device_uniqueid;
 	char *csd_token;
 	char *csd_ticket;
 	char *csd_stuburl;
@@ -152,14 +194,35 @@ struct openconnect_info {
 	char *csd_scriptname;
 	xmlNode *opaque_srvdata;
 
+	char *profile_url;
+	char *profile_sha1;
+
 #ifdef LIBPROXY_HDR
 	pxProxyFactory *proxy_factory;
 #endif
 	char *proxy_type;
 	char *proxy;
 	int proxy_port;
+	int proxy_fd;
+	char *proxy_user;
+	char *proxy_pass;
+	struct proxy_auth_state auth[MAX_AUTH_TYPES];
+#ifdef HAVE_GSSAPI
+	gss_name_t gss_target_name;
+	gss_ctx_id_t gss_context;
+#endif
+#ifdef _WIN32
+	CredHandle ntlm_sspi_cred;
+	CtxtHandle ntlm_sspi_ctx;
+	CredHandle sspi_cred;
+	CtxtHandle sspi_ctx;
+	char *sspi_target_name;
+#else
+	int ntlm_helper_fd;
+#endif
+	int authmethods_set;
 
-	const char *localname;
+	char *localname;
 	char *hostname;
 	char *unique_hostname;
 	int port;
@@ -169,8 +232,8 @@ struct openconnect_info {
 	const char *sslkey;
 	int cert_type;
 	char *cert_password;
-	const char *cafile;
-	const char *servercert;
+	char *cafile;
+	char *servercert;
 	const char *xmlconfig;
 	char xmlsha1[(SHA1_SIZE * 2) + 1];
 	char *authgroup;
@@ -199,10 +262,11 @@ struct openconnect_info {
 	OPENCONNECT_X509 *peer_cert;
 
 	char *cookie; /* Pointer to within cookies list */
-	struct vpn_option *cookies;
-	struct vpn_option *cstp_options;
-	struct vpn_option *dtls_options;
+	struct oc_vpn_option *cookies;
+	struct oc_vpn_option *cstp_options;
+	struct oc_vpn_option *dtls_options;
 
+	unsigned pfs;
 #if defined(OPENCONNECT_OPENSSL)
 	X509 *cert_x509;
 	SSL_CTX *https_ctx;
@@ -247,60 +311,65 @@ struct openconnect_info {
 #if defined(DTLS_OPENSSL)
 	SSL_CTX *dtls_ctx;
 	SSL *dtls_ssl;
-	SSL *new_dtls_ssl;
 	SSL_SESSION *dtls_session;
 #elif defined(DTLS_GNUTLS)
-	/* Call these *_ssl rather than *_sess because they're just
-	   pointers, and generic code (in mainloop.c for example)
-	   wants to check if they're NULL or not. No point in being
-	   differently named to the OpenSSL variant, and forcing us to
-	   have ifdefs or accessor macros for them. */
+	/* Call this dtls_ssl rather than dtls_sess because it's just a
+	   pointer, and generic code in dtls.c wants to check if it's
+	   NULL or not or pass it to DTLS_SEND/DTLS_RECV. This way we
+	   have fewer ifdefs and accessor macros for it. */
 	gnutls_session_t dtls_ssl;
-	gnutls_session_t new_dtls_ssl;
 #endif
+	int dtls_state;
 	struct keepalive_info dtls_times;
 	unsigned char dtls_session_id[32];
 	unsigned char dtls_secret[48];
 
 	char *dtls_cipher;
-	const char *vpnc_script;
+	char *vpnc_script;
 	int script_tun;
 	char *ifname;
 
-	int actual_mtu;
 	int reqmtu, basemtu;
 	const char *banner;
-	const char *vpn_addr;
-	const char *vpn_netmask;
-	const char *vpn_addr6;
-	const char *vpn_netmask6;
-	const char *vpn_dns[3];
-	const char *vpn_nbns[3];
-	const char *vpn_domain;
-	const char *vpn_proxy_pac;
-	struct split_include *split_dns;
-	struct split_include *split_includes;
-	struct split_include *split_excludes;
 
-	int select_nfds;
-	fd_set select_rfds;
-	fd_set select_wfds;
-	fd_set select_efds;
+	struct oc_ip_info ip_info;
+
+#ifdef _WIN32
+	long dtls_monitored, ssl_monitored, cmd_monitored, tun_monitored;
+	HANDLE dtls_event, ssl_event, cmd_event;
+#else
+	int _select_nfds;
+	fd_set _select_rfds;
+	fd_set _select_wfds;
+	fd_set _select_efds;
+#endif
 
 #ifdef __sun__
 	int ip_fd;
 	int ip6_fd;
 #endif
+#ifdef _WIN32
+	HANDLE tun_fh;
+	OVERLAPPED tun_rd_overlap, tun_wr_overlap;
+	int tun_rd_pending;
+#else
 	int tun_fd;
+#endif
 	int ssl_fd;
 	int dtls_fd;
-	int new_dtls_fd;
-	int cancel_fd;
+
+	int cmd_fd;
+	int cmd_fd_write;
+	int got_cancel_cmd;
+	int got_pause_cmd;
+	char cancel_type;
 
 	struct pkt *incoming_queue;
 	struct pkt *outgoing_queue;
 	int outgoing_qlen;
 	int max_qlen;
+	struct oc_stats stats;
+	openconnect_stats_vfn stats_handler;
 
 	socklen_t peer_addrlen;
 	struct sockaddr *peer_addr;
@@ -318,8 +387,40 @@ struct openconnect_info {
 	openconnect_write_new_config_vfn write_new_config;
 	openconnect_process_auth_form_vfn process_auth_form;
 	openconnect_progress_vfn progress;
+	openconnect_protect_socket_vfn protect_socket;
+
+	int (*ssl_read)(struct openconnect_info *vpninfo, char *buf, size_t len);
+	int (*ssl_gets)(struct openconnect_info *vpninfo, char *buf, size_t len);
+	int (*ssl_write)(struct openconnect_info *vpninfo, char *buf, size_t len);
 };
 
+#ifdef _WIN32
+#define monitor_read_fd(_v, _n) _v->_n##_monitored |= FD_READ
+#define monitor_write_fd(_v, _n) _v->_n##_monitored |= FD_WRITE
+#define monitor_except_fd(_v, _n) _v->_n##_monitored |= FD_CLOSE
+#define unmonitor_read_fd(_v, _n) _v->_n##_monitored &= ~FD_READ
+#define unmonitor_write_fd(_v, _n) _v->_n##_monitored &= ~FD_WRITE
+#define unmonitor_except_fd(_v, _n) _v->_n##_monitored &= ~FD_CLOSE
+
+#define monitor_fd_new(_v, _n) do { if (!_v->_n##_event) _v->_n##_event = CreateEvent(NULL, FALSE, FALSE, NULL); } while (0)
+#define read_fd_monitored(_v, _n) (_v->_n##_monitored & FD_READ)
+
+#else
+#define monitor_read_fd(_v, _n) FD_SET(_v-> _n##_fd, &vpninfo->_select_rfds)
+#define unmonitor_read_fd(_v, _n) FD_CLR(_v-> _n##_fd, &vpninfo->_select_rfds)
+#define monitor_write_fd(_v, _n) FD_SET(_v-> _n##_fd, &vpninfo->_select_wfds)
+#define unmonitor_write_fd(_v, _n) FD_CLR(_v-> _n##_fd, &vpninfo->_select_wfds)
+#define monitor_except_fd(_v, _n) FD_SET(_v-> _n##_fd, &vpninfo->_select_efds)
+#define unmonitor_except_fd(_v, _n) FD_CLR(_v-> _n##_fd, &vpninfo->_select_efds)
+
+#define monitor_fd_new(_v, _n) do { \
+		if (_v->_select_nfds <= vpninfo->_n##_fd) \
+			vpninfo->_select_nfds = vpninfo->_n##_fd + 1; \
+	} while (0)
+
+#define read_fd_monitored(_v, _n) FD_ISSET(_v->_n##_fd, &_v->_select_rfds)
+#endif
+
 #if (defined(DTLS_OPENSSL) && defined(SSL_OP_CISCO_ANYCONNECT)) || \
     (defined(DTLS_GNUTLS) && defined(HAVE_GNUTLS_SESSION_SET_PREMASTER))
 #define HAVE_DTLS 1
@@ -339,7 +440,7 @@ struct openconnect_info {
 
 /****************************************************************************/
 /* Oh Solaris how we hate thee! */
-#ifdef __sun__
+#ifdef HAVE_SUNOS_BROKEN_TIME
 #define time(x) openconnect__time(x)
 time_t openconnect__time(time_t *t);
 #endif
@@ -355,6 +456,53 @@ ssize_t openconnect__getline(char **lineptr, size_t *n, FILE *stream);
 #define strcasestr openconnect__strcasestr
 char *openconnect__strcasestr(const char *haystack, const char *needle);
 #endif
+#ifndef HAVE_STRNDUP
+#undef strndup
+#define strndup openconnect__strndup
+char *openconnect__strndup(const char *s, size_t n);
+#endif
+#ifndef HAVE_SETENV
+#define setenv openconnect__setenv
+int openconnect__setenv(const char *name, const char *value, int overwrite);
+#endif
+#ifndef HAVE_UNSETENV
+#define unsetenv openconnect__unsetenv
+void openconnect__unsetenv(const char *name);
+#endif
+
+#ifndef HAVE_INET_ATON
+#define inet_aton openconnect__inet_aton
+int openconnect__inet_aton(const char *cp, struct in_addr *addr);
+#endif
+
+static inline int set_sock_nonblock(int fd)
+{
+#ifdef _WIN32
+	unsigned long mode = 1;
+	return ioctlsocket(fd, FIONBIO, &mode);
+#else
+	return fcntl(fd, F_SETFL, fcntl(fd, F_GETFL) | O_NONBLOCK);
+#endif
+}
+static inline int set_fd_cloexec(int fd)
+{
+#ifdef _WIN32
+	return 0; /* Windows has O_INHERIT but... */
+#else
+	return fcntl(fd, F_SETFD, fcntl(fd, F_GETFD) | FD_CLOEXEC);
+#endif
+}
+
+#ifdef _WIN32
+#define pipe(fds) _pipe(fds, 4096, O_BINARY)
+void openconnect__win32_sock_init();
+int dumb_socketpair(int socks[2], int make_overlapped);
+#else
+#define closesocket close
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
+#endif
 
 /* I always coded as if it worked like this. Now it does. */
 #define realloc_inplace(p, size) do {			\
@@ -366,26 +514,33 @@ char *openconnect__strcasestr(const char *haystack, const char *needle);
 
 /****************************************************************************/
 
-/* tun.c */
-int setup_tun(struct openconnect_info *vpninfo);
-int tun_mainloop(struct openconnect_info *vpninfo, int *timeout);
-void shutdown_tun(struct openconnect_info *vpninfo);
+/* script.c */
+int setenv_int(const char *opt, int value);
+void set_script_env(struct openconnect_info *vpninfo);
 int script_config_tun(struct openconnect_info *vpninfo, const char *reason);
 
+/* tun.c / tun-win32.c */
+void os_shutdown_tun(struct openconnect_info *vpninfo);
+int os_read_tun(struct openconnect_info *vpninfo, struct pkt *pkt);
+int os_write_tun(struct openconnect_info *vpninfo, struct pkt *pkt);
+intptr_t os_setup_tun(struct openconnect_info *vpninfo);
+
 /* dtls.c */
 unsigned char unhex(const char *data);
-int setup_dtls(struct openconnect_info *vpninfo);
 int dtls_mainloop(struct openconnect_info *vpninfo, int *timeout);
 int dtls_try_handshake(struct openconnect_info *vpninfo);
 int connect_dtls_socket(struct openconnect_info *vpninfo);
+void dtls_close(struct openconnect_info *vpninfo);
+void dtls_shutdown(struct openconnect_info *vpninfo);
+int dtls_reconnect(struct openconnect_info *vpninfo);
 
 /* cstp.c */
-int make_cstp_connection(struct openconnect_info *vpninfo);
 int cstp_mainloop(struct openconnect_info *vpninfo, int *timeout);
 int cstp_bye(struct openconnect_info *vpninfo, const char *reason);
-int cstp_reconnect(struct openconnect_info *vpninfo);
+void cstp_free_splits(struct openconnect_info *vpninfo);
 
 /* ssl.c */
+unsigned string_is_hostname(const char* str);
 int connect_https_socket(struct openconnect_info *vpninfo);
 int request_passphrase(struct openconnect_info *vpninfo, const char *label,
 		       char **response, const char *fmt, ...);
@@ -400,16 +555,19 @@ int openconnect_print_err_cb(const char *str, size_t len, void *ptr);
 const char *keystore_strerror(int err);
 int keystore_fetch(const char *key, unsigned char **result);
 #endif
+void cmd_fd_set(struct openconnect_info *vpninfo, fd_set *fds, int *maxfd);
+void check_cmd_fd(struct openconnect_info *vpninfo, fd_set *fds);
+int is_cancel_pending(struct openconnect_info *vpninfo, fd_set *fds);
+void poll_cmd_fd(struct openconnect_info *vpninfo, int timeout);
 
-/* ${SSL_LIBRARY}.c */
-int openconnect_SSL_gets(struct openconnect_info *vpninfo, char *buf, size_t len);
-int openconnect_SSL_write(struct openconnect_info *vpninfo, char *buf, size_t len);
-int openconnect_SSL_read(struct openconnect_info *vpninfo, char *buf, size_t len);
+/* {gnutls,openssl}.c */
 int openconnect_open_https(struct openconnect_info *vpninfo);
 void openconnect_close_https(struct openconnect_info *vpninfo, int final);
+int cstp_handshake(struct openconnect_info *vpninfo, unsigned init);
 int get_cert_md5_fingerprint(struct openconnect_info *vpninfo, OPENCONNECT_X509 *cert,
 			     char *buf);
 int openconnect_sha1(unsigned char *result, void *data, int len);
+int openconnect_md5(unsigned char *result, void *data, int len);
 int openconnect_random(void *bytes, int len);
 int openconnect_local_cert_md5(struct openconnect_info *vpninfo,
 			       char *buf);
@@ -420,21 +578,20 @@ int openconnect_local_cert_md5(struct openconnect_info *vpninfo,
 #endif
 
 /* mainloop.c */
-int vpn_add_pollfd(struct openconnect_info *vpninfo, int fd, short events);
-int vpn_mainloop(struct openconnect_info *vpninfo);
+int tun_mainloop(struct openconnect_info *vpninfo, int *timeout);
 int queue_new_packet(struct pkt **q, void *buf, int len);
 void queue_packet(struct pkt **q, struct pkt *new);
 int keepalive_action(struct keepalive_info *ka, int *timeout);
 int ka_stalled_action(struct keepalive_info *ka, int *timeout);
 
-extern int killed;
-
 /* xml.c */
 int config_lookup_host(struct openconnect_info *vpninfo, const char *host);
 
 /* auth.c */
 int parse_xml_response(struct openconnect_info *vpninfo, char *response,
 		       struct oc_auth_form **form, int *cert_rq);
+int process_auth_form(struct openconnect_info *vpninfo,
+		      struct oc_auth_form *form);
 int handle_auth_form(struct openconnect_info *vpninfo, struct oc_auth_form *form,
 		     char *request_body, int req_len, const char **method,
 		     const char **request_body_type);
@@ -443,13 +600,31 @@ int xmlpost_initial_req(struct openconnect_info *vpninfo, char *request_body, in
 int prepare_stoken(struct openconnect_info *vpninfo);
 
 /* http.c */
+struct oc_text_buf *buf_alloc(void);
+int buf_ensure_space(struct oc_text_buf *buf, int len);
+void  __attribute__ ((format (printf, 2, 3)))
+	buf_append(struct oc_text_buf *buf, const char *fmt, ...);
+void buf_append_bytes(struct oc_text_buf *buf, const void *bytes, int len);
+void buf_append_base64(struct oc_text_buf *buf, const void *bytes, int len);
+int openconnect_base64_decode(void **out, const char *in);
+int buf_error(struct oc_text_buf *buf);
+int buf_free(struct oc_text_buf *buf);
 char *openconnect_create_useragent(const char *base);
 int process_proxy(struct openconnect_info *vpninfo, int ssl_sock);
 int internal_parse_url(char *url, char **res_proto, char **res_host,
 		       int *res_port, char **res_path, int default_port);
 
-/* ssl_ui.c */
-int set_openssl_ui(void);
+/* ntlm.c */
+int ntlm_authorization(struct openconnect_info *vpninfo, struct oc_text_buf *buf);
+void cleanup_ntlm_auth(struct openconnect_info *vpninfo);
+
+/* gssapi.c */
+int gssapi_authorization(struct openconnect_info *vpninfo, struct oc_text_buf *buf);
+void cleanup_gssapi_auth(struct openconnect_info *vpninfo);
+int socks_gssapi_auth(struct openconnect_info *vpninfo);
+
+/* digest.c */
+int digest_authorization(struct openconnect_info *vpninfo, struct oc_text_buf *buf);
 
 /* version.c */
 extern const char *openconnect_version_str;
