@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2012 Intel Corporation.
+ * Copyright © 2008-2014 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -13,22 +13,17 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to:
- *
- *   Free Software Foundation, Inc.
- *   51 Franklin Street, Fifth Floor,
- *   Boston, MA 02110-1301 USA
  */
 
+#include <config.h>
+
 #include <errno.h>
 #include <string.h>
 #include <sys/types.h>
-#include <netinet/in.h>
-#include <arpa/inet.h>
 #include <ctype.h>
 
+#include "openconnect-internal.h"
+
 #include <openssl/ssl.h>
 #include <openssl/err.h>
 #include <openssl/engine.h>
@@ -39,7 +34,6 @@
 #include <openssl/x509.h>
 #include <openssl/bio.h>
 
-#include "openconnect-internal.h"
 
 int openconnect_sha1(unsigned char *result, void *data, int len)
 {
@@ -52,6 +46,17 @@ int openconnect_sha1(unsigned char *result, void *data, int len)
 	return 0;
 }
 
+int openconnect_md5(unsigned char *result, void *data, int len)
+{
+	EVP_MD_CTX c;
+
+	EVP_MD_CTX_init(&c);
+	EVP_Digest(data, len, result, NULL, EVP_md5(), NULL);
+	EVP_MD_CTX_cleanup(&c);
+
+	return 0;
+}
+
 int openconnect_get_cert_DER(struct openconnect_info *vpninfo,
 			     OPENCONNECT_X509 *cert, unsigned char **buf)
 {
@@ -86,7 +91,7 @@ int openconnect_random(void *bytes, int len)
 /* Helper functions for reading/writing lines over SSL.
    We could use cURL for the HTTP stuff, but it's overkill */
 
-int openconnect_SSL_write(struct openconnect_info *vpninfo, char *buf, size_t len)
+static int openconnect_openssl_write(struct openconnect_info *vpninfo, char *buf, size_t len)
 {
 	size_t orig_len = len;
 
@@ -112,14 +117,9 @@ int openconnect_SSL_write(struct openconnect_info *vpninfo, char *buf, size_t le
 				openconnect_report_ssl_errors(vpninfo);
 				return -EIO;
 			}
-			if (vpninfo->cancel_fd != -1) {
-				FD_SET(vpninfo->cancel_fd, &rd_set);
-				if (vpninfo->cancel_fd > vpninfo->ssl_fd)
-					maxfd = vpninfo->cancel_fd;
-			}
+			cmd_fd_set(vpninfo, &rd_set, &maxfd);
 			select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
-			if (vpninfo->cancel_fd != -1 &&
-			    FD_ISSET(vpninfo->cancel_fd, &rd_set)) {
+			if (is_cancel_pending(vpninfo, &rd_set)) {
 				vpn_progress(vpninfo, PRG_ERR, _("SSL write cancelled\n"));
 				return -EINTR;
 			}
@@ -128,7 +128,7 @@ int openconnect_SSL_write(struct openconnect_info *vpninfo, char *buf, size_t le
 	return orig_len;
 }
 
-int openconnect_SSL_read(struct openconnect_info *vpninfo, char *buf, size_t len)
+static int openconnect_openssl_read(struct openconnect_info *vpninfo, char *buf, size_t len)
 {
 	int done;
 
@@ -149,14 +149,9 @@ int openconnect_SSL_read(struct openconnect_info *vpninfo, char *buf, size_t len
 			openconnect_report_ssl_errors(vpninfo);
 			return -EIO;
 		}
-		if (vpninfo->cancel_fd != -1) {
-			FD_SET(vpninfo->cancel_fd, &rd_set);
-			if (vpninfo->cancel_fd > vpninfo->ssl_fd)
-				maxfd = vpninfo->cancel_fd;
-		}
+		cmd_fd_set(vpninfo, &rd_set, &maxfd);
 		select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
-		if (vpninfo->cancel_fd != -1 &&
-		    FD_ISSET(vpninfo->cancel_fd, &rd_set)) {
+		if (is_cancel_pending(vpninfo, &rd_set)) {
 			vpn_progress(vpninfo, PRG_ERR, _("SSL read cancelled\n"));
 			return -EINTR;
 		}
@@ -164,7 +159,7 @@ int openconnect_SSL_read(struct openconnect_info *vpninfo, char *buf, size_t len
 	return done;
 }
 
-int openconnect_SSL_gets(struct openconnect_info *vpninfo, char *buf, size_t len)
+static int openconnect_openssl_gets(struct openconnect_info *vpninfo, char *buf, size_t len)
 {
 	int i = 0;
 	int ret;
@@ -207,14 +202,9 @@ int openconnect_SSL_gets(struct openconnect_info *vpninfo, char *buf, size_t len
 				ret = -EIO;
 				break;
 			}
-			if (vpninfo->cancel_fd != -1) {
-				FD_SET(vpninfo->cancel_fd, &rd_set);
-				if (vpninfo->cancel_fd > vpninfo->ssl_fd)
-					maxfd = vpninfo->cancel_fd;
-			}
+			cmd_fd_set(vpninfo, &rd_set, &maxfd);
 			select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
-			if (vpninfo->cancel_fd != -1 &&
-			    FD_ISSET(vpninfo->cancel_fd, &rd_set)) {
+			if (is_cancel_pending(vpninfo, &rd_set)) {
 				vpn_progress(vpninfo, PRG_ERR, _("SSL read cancelled\n"));
 				ret = -EINTR;
 				break;
@@ -306,7 +296,7 @@ static int ui_flush(UI *ui)
 	struct ui_form_opt *opt;
 	int ret;
 
-	ret = vpninfo->process_auth_form(vpninfo->cbdata, &ui_data->form);
+	ret = process_auth_form(vpninfo, &ui_data->form);
 	if (ret)
 		return 0;
 
@@ -426,26 +416,27 @@ static int load_pkcs12_certificate(struct openconnect_info *vpninfo, PKCS12 *p12
 	   when PKCS12_parse() returns an error, but *ca is left pointing
 	   to the freed memory. */
 	ca = NULL;
-	if (!pass && request_passphrase(vpninfo, "openconnect_pkcs12", &pass,
-					_("Enter PKCS#12 pass phrase:")) < 0) {
-		PKCS12_free(p12);
-		return -EINVAL;
-	}
 	if (!PKCS12_parse(p12, pass, &pkey, &cert, &ca)) {
 		unsigned long err = ERR_peek_error();
 
-		openconnect_report_ssl_errors(vpninfo);
-
 		if (ERR_GET_LIB(err) == ERR_LIB_PKCS12 &&
 		    ERR_GET_FUNC(err) == PKCS12_F_PKCS12_PARSE &&
 		    ERR_GET_REASON(err) == PKCS12_R_MAC_VERIFY_FAILURE) {
-			vpn_progress(vpninfo, PRG_ERR,
-				     _("Parse PKCS#12 failed (wrong passphrase?)\n"));
+			if (pass)
+				vpn_progress(vpninfo, PRG_ERR,
+					     _("Failed to decrypt PKCS#12 certificate file\n"));
 			free(pass);
-			pass = NULL;
+			if (request_passphrase(vpninfo, "openconnect_pkcs12", &pass,
+					       _("Enter PKCS#12 pass phrase:")) < 0) {
+				PKCS12_free(p12);
+				return -EINVAL;
+			}
+
 			goto retrypass;
 		}
 
+		openconnect_report_ssl_errors(vpninfo);
+
 		vpn_progress(vpninfo, PRG_ERR,
 			     _("Parse PKCS#12 failed (see above errors)\n"));
 		PKCS12_free(p12);
@@ -669,7 +660,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		return -EINVAL;
 	}
 
-	vpn_progress(vpninfo, PRG_TRACE,
+	vpn_progress(vpninfo, PRG_DEBUG,
 		     _("Using certificate file %s\n"), vpninfo->cert);
 
 	if (strncmp(vpninfo->cert, "keystore:", 9) &&
@@ -678,7 +669,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 		FILE *f;
 		PKCS12 *p12;
 
-		f = fopen(vpninfo->cert, "r");
+		f = fopen(vpninfo->cert, "rb");
 		if (!f) {
 			vpn_progress(vpninfo, PRG_ERR,
 				     _("Failed to open certificate file %s: %s\n"),
@@ -766,7 +757,7 @@ static int load_certificate(struct openconnect_info *vpninfo)
 #endif /* ANDROID_KEYSTORE */
 
 	if (vpninfo->cert_type == CERT_TYPE_UNKNOWN) {
-		FILE *f = fopen(vpninfo->sslkey, "r");
+		FILE *f = fopen(vpninfo->sslkey, "rb");
 		char buf[256];
 
 		if (!f) {
@@ -783,7 +774,8 @@ static int load_certificate(struct openconnect_info *vpninfo)
 				break;
 			} else if (!strcmp(buf, "-----BEGIN RSA PRIVATE KEY-----\n") ||
 				   !strcmp(buf, "-----BEGIN DSA PRIVATE KEY-----\n") ||
-				   !strcmp(buf, "-----BEGIN ENCRYPTED PRIVATE KEY-----\n")) {
+				   !strcmp(buf, "-----BEGIN ENCRYPTED PRIVATE KEY-----\n") ||
+				   !strcmp(buf, "-----BEGIN PRIVATE KEY-----\n")) {
 				vpninfo->cert_type = CERT_TYPE_PEM;
 				break;
 			}
@@ -971,14 +963,14 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 				continue;
 
 			if (!match_hostname(vpninfo->hostname, str)) {
-				vpn_progress(vpninfo, PRG_TRACE,
+				vpn_progress(vpninfo, PRG_DEBUG,
 					     _("Matched DNS altname '%s'\n"),
 					     str);
 				GENERAL_NAMES_free(altnames);
 				OPENSSL_free(str);
 				return 0;
 			} else {
-				vpn_progress(vpninfo, PRG_TRACE,
+				vpn_progress(vpninfo, PRG_DEBUG,
 					     _("No match for altname '%s'\n"),
 					     str);
 			}
@@ -1003,14 +995,14 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 
 			if (this->d.ip->length == addrlen &&
 			    !memcmp(addrbuf, this->d.ip->data, addrlen)) {
-				vpn_progress(vpninfo, PRG_TRACE,
+				vpn_progress(vpninfo, PRG_DEBUG,
 					     _("Matched %s address '%s'\n"),
 					     (family == AF_INET6) ? "IPv6" : "IPv4",
 					     host);
 				GENERAL_NAMES_free(altnames);
 				return 0;
 			} else {
-				vpn_progress(vpninfo, PRG_TRACE,
+				vpn_progress(vpninfo, PRG_DEBUG,
 					     _("No match for %s address '%s'\n"),
 					     (family == AF_INET6) ? "IPv6" : "IPv4",
 					     host);
@@ -1052,12 +1044,12 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 				goto no_uri_match;
 
 			if (url_path) {
-				vpn_progress(vpninfo, PRG_TRACE,
+				vpn_progress(vpninfo, PRG_DEBUG,
 					     _("URI '%s' has non-empty path; ignoring\n"),
 					     str);
 				goto no_uri_match_silent;
 			}
-			vpn_progress(vpninfo, PRG_TRACE,
+			vpn_progress(vpninfo, PRG_DEBUG,
 				     _("Matched URI '%s'\n"),
 				     str);
 			free(url_proto);
@@ -1068,7 +1060,7 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 			return 0;
 
 		no_uri_match:
-			vpn_progress(vpninfo, PRG_TRACE,
+			vpn_progress(vpninfo, PRG_DEBUG,
 				     _("No match for URI '%s'\n"),
 				     str);
 		no_uri_match_silent:
@@ -1123,7 +1115,7 @@ static int match_cert_hostname(struct openconnect_info *vpninfo, X509 *peer_cert
 			     subjstr, vpninfo->hostname);
 		ret = -EINVAL;
 	} else {
-		vpn_progress(vpninfo, PRG_TRACE,
+		vpn_progress(vpninfo, PRG_DEBUG,
 			     _("Matched peer certificate subject name '%s'\n"),
 			     subjstr);
 	}
@@ -1300,7 +1292,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 					     _("Loading certificate failed. Aborting.\n"));
 				SSL_CTX_free(vpninfo->https_ctx);
 				vpninfo->https_ctx = NULL;
-				close(ssl_sock);
+				closesocket(ssl_sock);
 				return err;
 			}
 			check_certificate_expiry(vpninfo);
@@ -1319,6 +1311,9 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 						 ssl_app_verify_callback, NULL);
 #endif
 		SSL_CTX_set_default_verify_paths(vpninfo->https_ctx);
+		
+		if (vpninfo->pfs)
+			SSL_CTX_set_cipher_list(vpninfo->https_ctx, "HIGH:!aNULL:!eNULL:-RSA");
 
 #ifdef ANDROID_KEYSTORE
 		if (vpninfo->cafile && !strncmp(vpninfo->cafile, "keystore:", 9)) {
@@ -1330,7 +1325,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 			if (!b) {
 				SSL_CTX_free(vpninfo->https_ctx);
 				vpninfo->https_ctx = NULL;
-				close(ssl_sock);
+				closesocket(ssl_sock);
 				return -EINVAL;
 			}
 
@@ -1344,7 +1339,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 				openconnect_report_ssl_errors(vpninfo);
 				SSL_CTX_free(vpninfo->https_ctx);
 				vpninfo->https_ctx = NULL;
-				close(ssl_sock);
+				closesocket(ssl_sock);
 				return -ENOENT;
 			}
 
@@ -1368,7 +1363,7 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 				openconnect_report_ssl_errors(vpninfo);
 				SSL_CTX_free(vpninfo->https_ctx);
 				vpninfo->https_ctx = NULL;
-				close(ssl_sock);
+				closesocket(ssl_sock);
 				return -EINVAL;
 			}
 		}
@@ -1381,6 +1376,14 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 	BIO_set_nbio(https_bio, 1);
 	SSL_set_bio(https_ssl, https_bio, https_bio);
 
+#if 0
+	/* Should be enabled on openssl versions that support
+	   SSL_set_tlsext_host_name() after the F5 firewall workaround
+	   is enabled */
+	if (string_is_hostname(vpninfo->hostname))
+		SSL_set_tlsext_host_name(https_ssl, vpninfo->hostname);
+#endif
+
 	vpn_progress(vpninfo, PRG_INFO, _("SSL negotiation with %s\n"),
 		     vpninfo->hostname);
 
@@ -1400,34 +1403,33 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 			vpn_progress(vpninfo, PRG_ERR, _("SSL connection failure\n"));
 			openconnect_report_ssl_errors(vpninfo);
 			SSL_free(https_ssl);
-			close(ssl_sock);
+			closesocket(ssl_sock);
 			return -EINVAL;
 		}
 
-		if (vpninfo->cancel_fd != -1) {
-			FD_SET(vpninfo->cancel_fd, &rd_set);
-			if (vpninfo->cancel_fd > ssl_sock)
-				maxfd = vpninfo->cancel_fd;
-		}
+		cmd_fd_set(vpninfo, &rd_set, &maxfd);
 		select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
-		if (vpninfo->cancel_fd != -1 &&
-		    FD_ISSET(vpninfo->cancel_fd, &rd_set)) {
+		if (is_cancel_pending(vpninfo, &rd_set)) {
 			vpn_progress(vpninfo, PRG_ERR, _("SSL connection cancelled\n"));
 			SSL_free(https_ssl);
-			close(ssl_sock);
+			closesocket(ssl_sock);
 			return -EINVAL;
 		}
 	}
 
 	if (verify_peer(vpninfo, https_ssl)) {
 		SSL_free(https_ssl);
-		close(ssl_sock);
+		closesocket(ssl_sock);
 		return -EINVAL;
 	}
 
 	vpninfo->ssl_fd = ssl_sock;
 	vpninfo->https_ssl = https_ssl;
 
+	vpninfo->ssl_read = openconnect_openssl_read;
+	vpninfo->ssl_write = openconnect_openssl_write;
+	vpninfo->ssl_gets = openconnect_openssl_gets;
+
 	/* Stash this now, because it might not be available later if the
 	   server has disconnected. */
 	vpninfo->peer_cert = SSL_get_peer_certificate(vpninfo->https_ssl);
@@ -1438,6 +1440,11 @@ int openconnect_open_https(struct openconnect_info *vpninfo)
 	return 0;
 }
 
+int cstp_handshake(struct openconnect_info *vpninfo, unsigned init)
+{
+	return -EOPNOTSUPP;
+}
+
 void openconnect_close_https(struct openconnect_info *vpninfo, int final)
 {
 	if (vpninfo->peer_cert) {
@@ -1449,10 +1456,10 @@ void openconnect_close_https(struct openconnect_info *vpninfo, int final)
 		vpninfo->https_ssl = NULL;
 	}
 	if (vpninfo->ssl_fd != -1) {
-		close(vpninfo->ssl_fd);
-		FD_CLR(vpninfo->ssl_fd, &vpninfo->select_rfds);
-		FD_CLR(vpninfo->ssl_fd, &vpninfo->select_wfds);
-		FD_CLR(vpninfo->ssl_fd, &vpninfo->select_efds);
+		closesocket(vpninfo->ssl_fd);
+		unmonitor_read_fd(vpninfo, ssl);
+		unmonitor_write_fd(vpninfo, ssl);
+		unmonitor_except_fd(vpninfo, ssl);
 		vpninfo->ssl_fd = -1;
 	}
 	if (final) {
@@ -1469,6 +1476,9 @@ void openconnect_close_https(struct openconnect_info *vpninfo, int final)
 
 void openconnect_init_ssl(void)
 {
+#ifdef _WIN32
+	openconnect__win32_sock_init();
+#endif
 	SSL_library_init();
 	ERR_clear_error();
 	SSL_load_error_strings();
