@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2012 Intel Corporation.
+ * Copyright © 2008-2014 Intel Corporation.
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
  *
@@ -13,21 +13,12 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to:
- *
- *   Free Software Foundation, Inc.
- *   51 Franklin Street, Fifth Floor,
- *   Boston, MA 02110-1301 USA
  */
 
+#include <config.h>
+
 #include <sys/types.h>
-#include <sys/socket.h>
 #include <sys/stat.h>
-#include <netinet/in.h>
-#include <arpa/inet.h>
-#include <netdb.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <string.h>
@@ -35,6 +26,7 @@
 #include <errno.h>
 #include <stdlib.h>
 #include <stdarg.h>
+#include <time.h>
 #if defined(__linux__) || defined(__ANDROID__)
 #include <sys/vfs.h>
 #elif defined(__FreeBSD__) || defined(__FreeBSD_kernel__) || defined(__OpenBSD__) || defined(__APPLE__)
@@ -57,6 +49,14 @@
 #define AI_NUMERICSERV 0
 #endif
 
+static inline int connect_pending()
+{
+#ifdef _WIN32
+	return WSAGetLastError() == WSAEWOULDBLOCK;
+#else
+	return errno == EINPROGRESS;
+#endif
+}
 static int cancellable_connect(struct openconnect_info *vpninfo, int sockfd,
 			       const struct sockaddr *addr, socklen_t addrlen)
 {
@@ -65,34 +65,47 @@ static int cancellable_connect(struct openconnect_info *vpninfo, int sockfd,
 	fd_set wr_set, rd_set;
 	int maxfd = sockfd;
 
-	fcntl(sockfd, F_SETFL, fcntl(sockfd, F_GETFL) | O_NONBLOCK);
+	set_sock_nonblock(sockfd);
+	if (vpninfo->protect_socket)
+		vpninfo->protect_socket(vpninfo->cbdata, sockfd);
 
-	if (connect(sockfd, addr, addrlen) < 0 && errno != EINPROGRESS)
+	if (connect(sockfd, addr, addrlen) < 0 && !connect_pending())
 		return -1;
 
-	FD_ZERO(&wr_set);
-	FD_ZERO(&rd_set);
-	FD_SET(sockfd, &wr_set);
-	if (vpninfo->cancel_fd != -1) {
-		FD_SET(vpninfo->cancel_fd, &rd_set);
-		if (vpninfo->cancel_fd > sockfd)
-			maxfd = vpninfo->cancel_fd;
-	}
+	do {
+		FD_ZERO(&wr_set);
+		FD_ZERO(&rd_set);
+		FD_SET(sockfd, &wr_set);
+		cmd_fd_set(vpninfo, &rd_set, &maxfd);
 
-	/* Later we'll render this whole exercise non-pointless by
-	   including a 'cancelfd' here too. */
-	select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
-	if (vpninfo->cancel_fd != -1 && FD_ISSET(vpninfo->cancel_fd, &rd_set)) {
-		vpn_progress(vpninfo, PRG_ERR, _("Socket connect cancelled\n"));
-		errno = EINTR;
-		return -1;
-	}
+		select(maxfd + 1, &rd_set, &wr_set, NULL, NULL);
+		if (is_cancel_pending(vpninfo, &rd_set)) {
+			vpn_progress(vpninfo, PRG_ERR, _("Socket connect cancelled\n"));
+			errno = EINTR;
+			return -1;
+		}
+	} while (!FD_ISSET(sockfd, &wr_set) && !vpninfo->got_pause_cmd);
 
 	/* Check whether connect() succeeded or failed by using
 	   getpeername(). See http://cr.yp.to/docs/connect.html */
 	return getpeername(sockfd, (void *)&peer, &peerlen);
 }
 
+/* checks whether the provided string is an IP or a hostname.
+ */
+unsigned string_is_hostname(const char *str)
+{
+	struct in_addr buf;
+
+	/* We don't use inet_pton() because an IPv6 literal is likely to
+	   be encased in []. So just check for a colon, which shouldn't
+	   occur in hostnames anyway. */
+	if (!str || inet_aton(str, &buf) || strchr(str, ':'))
+		return 0;
+
+	return 1;
+}
+
 int connect_https_socket(struct openconnect_info *vpninfo)
 {
 	int ssl_sock = -1;
@@ -102,6 +115,7 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 		vpninfo->port = 443;
 
 	if (vpninfo->peer_addr) {
+	reconnect:
 #ifdef SOCK_CLOEXEC
 		ssl_sock = socket(vpninfo->peer_addr->sa_family, SOCK_STREAM | SOCK_CLOEXEC, IPPROTO_IP);
 		if (ssl_sock < 0)
@@ -110,7 +124,7 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 			ssl_sock = socket(vpninfo->peer_addr->sa_family, SOCK_STREAM, IPPROTO_IP);
 			if (ssl_sock < 0)
 				goto reconn_err;
-			fcntl(ssl_sock, F_SETFD, fcntl(ssl_sock, F_GETFD) | FD_CLOEXEC);
+			set_fd_cloexec(ssl_sock);
 		}
 		if (cancellable_connect(vpninfo, ssl_sock, vpninfo->peer_addr, vpninfo->peer_addrlen)) {
 		reconn_err:
@@ -124,7 +138,7 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 					     vpninfo->hostname);
 			}
 			if (ssl_sock >= 0)
-				close(ssl_sock);
+				closesocket(ssl_sock);
 			return -EINVAL;
 		}
 	} else {
@@ -182,7 +196,7 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 			free(url);
 			free(proxies);
 			if (vpninfo->proxy)
-				vpn_progress(vpninfo, PRG_TRACE,
+				vpn_progress(vpninfo, PRG_DEBUG,
 					     _("Proxy from libproxy: %s://%s:%d/\n"),
 					     vpninfo->proxy_type, vpninfo->proxy, vpninfo->port);
 		}
@@ -196,15 +210,9 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 		}
 
 		if (hostname[0] == '[' && hostname[strlen(hostname)-1] == ']') {
-			/* Solaris has no strndup(). */
-			int len = strlen(hostname) - 2;
-			char *new_hostname = malloc(len + 1);
-			if (!new_hostname)
+			hostname = strndup(hostname + 1, strlen(hostname) - 2);
+			if (!hostname)
 				return -ENOMEM;
-			memcpy(new_hostname, hostname + 1, len);
-			new_hostname[len] = 0;
-
-			hostname = new_hostname;
 			hints.ai_flags |= AI_NUMERICHOST;
 		}
 
@@ -239,6 +247,7 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 					  rp->ai_protocol);
 			if (ssl_sock < 0)
 				continue;
+			set_fd_cloexec(ssl_sock);
 			if (cancellable_connect(vpninfo, ssl_sock, rp->ai_addr, rp->ai_addrlen) >= 0) {
 				/* Store the peer address we actually used, so that DTLS can
 				   use it again later */
@@ -246,7 +255,7 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 				if (!vpninfo->peer_addr) {
 					vpn_progress(vpninfo, PRG_ERR,
 						     _("Failed to allocate sockaddr storage\n"));
-					close(ssl_sock);
+					closesocket(ssl_sock);
 					return -ENOMEM;
 				}
 				vpninfo->peer_addrlen = rp->ai_addrlen;
@@ -270,7 +279,7 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 				}
 				break;
 			}
-			close(ssl_sock);
+			closesocket(ssl_sock);
 			ssl_sock = -1;
 		}
 		freeaddrinfo(result);
@@ -286,7 +295,13 @@ int connect_https_socket(struct openconnect_info *vpninfo)
 	if (vpninfo->proxy) {
 		err = process_proxy(vpninfo, ssl_sock);
 		if (err) {
-			close(ssl_sock);
+			closesocket(ssl_sock);
+			if (err == -EAGAIN) {
+				/* Proxy authentication failed and we need to retry */
+				vpn_progress(vpninfo, PRG_DEBUG,
+					     _("Reconnecting to proxy %s\n"), vpninfo->proxy);
+				goto reconnect;
+			}
 			return err;
 		}
 	}
@@ -305,7 +320,7 @@ int  __attribute__ ((format (printf, 2, 3)))
 	va_start(args, fmt);
 	vsnprintf(buf, 1023, fmt, args);
 	va_end(args);
-	return openconnect_SSL_write(vpninfo, buf, strlen(buf));
+	return vpninfo->ssl_write(vpninfo, buf, strlen(buf));
 
 }
 
@@ -318,9 +333,6 @@ int request_passphrase(struct openconnect_info *vpninfo, const char *label,
 	va_list args;
 	int ret;
 
-	if (!vpninfo->process_auth_form)
-		return -EINVAL;
-
 	buf[1023] = 0;
 	memset(&f, 0, sizeof(f));
 	va_start(args, fmt);
@@ -336,7 +348,7 @@ int request_passphrase(struct openconnect_info *vpninfo, const char *label,
 	o.label = buf;
 	o.value = NULL;
 
-	ret = vpninfo->process_auth_form(vpninfo->cbdata, &f);
+	ret = process_auth_form(vpninfo, &f);
 	if (!ret) {
 		*response = o.value;
 		return 0;
@@ -360,6 +372,31 @@ int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
 		return -ENOMEM;
 	return 0;
 }
+#elif defined(_WIN32)
+#include <fileapi.h>
+int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
+{
+	HANDLE h;
+	DWORD serial;
+	int success;
+
+	h = CreateFile(vpninfo->sslkey, 0, 0, NULL, OPEN_EXISTING,
+		       FILE_ATTRIBUTE_NORMAL, NULL);
+	if (h == INVALID_HANDLE_VALUE)
+		return -EIO;
+
+	success = GetVolumeInformationByHandleW(h, NULL, 0, &serial, NULL,
+						NULL, NULL, 0);
+	CloseHandle(h);
+
+	if (!success)
+		return -EIO;
+
+	if (asprintf(&vpninfo->cert_password, "%lx", serial))
+		return -ENOMEM;
+
+	return 0;
+}
 #else
 int openconnect_passphrase_from_fsid(struct openconnect_info *vpninfo)
 {
@@ -526,3 +563,68 @@ int keystore_fetch(const char *key, unsigned char **result)
 	return ret;
 }
 #endif
+
+void cmd_fd_set(struct openconnect_info *vpninfo, fd_set *fds, int *maxfd)
+{
+	if (vpninfo->cmd_fd != -1) {
+		FD_SET(vpninfo->cmd_fd, fds);
+		if (vpninfo->cmd_fd > *maxfd)
+			*maxfd = vpninfo->cmd_fd;
+	}
+}
+
+void check_cmd_fd(struct openconnect_info *vpninfo, fd_set *fds)
+{
+	char cmd;
+
+	if (vpninfo->cmd_fd == -1 || !FD_ISSET(vpninfo->cmd_fd, fds))
+		return;
+	if (vpninfo->cmd_fd_write == -1) {
+		/* legacy openconnect_set_cancel_fd() users */
+		vpninfo->got_cancel_cmd = 1;
+		return;
+	}
+
+	if (read(vpninfo->cmd_fd, &cmd, 1) != 1)
+		return;
+
+	switch (cmd) {
+	case OC_CMD_CANCEL:
+	case OC_CMD_DETACH:
+		vpninfo->got_cancel_cmd = 1;
+		vpninfo->cancel_type = cmd;
+		break;
+	case OC_CMD_PAUSE:
+		vpninfo->got_pause_cmd = 1;
+		break;
+	case OC_CMD_STATS:
+		if (vpninfo->stats_handler)
+			vpninfo->stats_handler(vpninfo->cbdata, &vpninfo->stats);
+	}
+}
+
+int is_cancel_pending(struct openconnect_info *vpninfo, fd_set *fds)
+{
+	check_cmd_fd(vpninfo, fds);
+	return vpninfo->got_cancel_cmd;
+}
+
+void poll_cmd_fd(struct openconnect_info *vpninfo, int timeout)
+{
+	fd_set rd_set;
+	int maxfd = 0;
+	time_t expiration = time(NULL) + timeout, now = 0;
+
+	while (now < expiration && !vpninfo->got_cancel_cmd && !vpninfo->got_pause_cmd) {
+		struct timeval tv;
+
+		now = time(NULL);
+		tv.tv_sec = now >= expiration ? 0 : expiration - now;
+		tv.tv_usec = 0;
+
+		FD_ZERO(&rd_set);
+		cmd_fd_set(vpninfo, &rd_set, &maxfd);
+		select(maxfd + 1, &rd_set, NULL, NULL, &tv);
+		check_cmd_fd(vpninfo, &rd_set);
+	}
+}
