@@ -1,8 +1,8 @@
 #!/bin/sh
 
-v="v5.02"
+v="v6.00"
 
-if [ -d .git ] && tag=`git describe --tags`; then
+if [ -d ${GIT_DIR:-.git} ] && tag=`git describe --tags`; then
 	v="$tag"
 
 	# Update the index from working tree first
