@@ -1,7 +1,7 @@
 /*
  * OpenConnect (SSL + DTLS) VPN client
  *
- * Copyright © 2008-2011 Intel Corporation.
+ * Copyright © 2008-2014 Intel Corporation.
  * Copyright © 2008 Nick Andrew <nick@nick-andrew.net>
  *
  * Author: David Woodhouse <dwmw2@infradead.org>
@@ -14,32 +14,77 @@
  * WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to:
- *
- *   Free Software Foundation, Inc.
- *   51 Franklin Street, Fifth Floor,
- *   Boston, MA 02110-1301 USA
  */
 
+#include <config.h>
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <sys/mman.h>
 #include <libxml/parser.h>
 #include <libxml/tree.h>
 #include <string.h>
+#include <errno.h>
 
 #include "openconnect-internal.h"
 
+static int load_xml_conf_file(struct openconnect_info *vpninfo, char **ptr,
+			      size_t *size)
+{
+	struct stat st;
+	int fd;
+	char *xmlfile = NULL;
+
+	fd = open(vpninfo->xmlconfig, O_RDONLY);
+	if (fd < 0) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to open XML config file: %s\n"),
+			     strerror(errno));
+		return 0;
+	}
+
+	if (fstat(fd, &st)) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to fstat() XML config file: %s\n"),
+			     strerror(errno));
+		goto err;
+	}
+
+	xmlfile = malloc(st.st_size);
+	if (!xmlfile) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to allocate %lu bytes for XML config file\n"),
+			     (unsigned long)st.st_size);
+		goto err;
+	}
+
+	if (read(fd, xmlfile, st.st_size) != st.st_size) {
+		vpn_progress(vpninfo, PRG_ERR,
+			     _("Failed to read XML config file: %s\n"),
+			     strerror(errno));
+		goto err;
+	}
+
+	*ptr = xmlfile;
+	*size = st.st_size;
+
+	close(fd);
+
+	return 1;
+
+err:
+	close(fd);
+	free(xmlfile);
+	return -1;
+}
+
 int config_lookup_host(struct openconnect_info *vpninfo, const char *host)
 {
-	int fd, i;
-	struct stat st;
+	int i, ret;
+	size_t size;
 	char *xmlfile;
 	unsigned char sha1[SHA1_SIZE];
 	xmlDocPtr xml_doc;
@@ -48,42 +93,29 @@ int config_lookup_host(struct openconnect_info *vpninfo, const char *host)
 	if (!vpninfo->xmlconfig)
 		return 0;
 
-	fd = open(vpninfo->xmlconfig, O_RDONLY);
-	if (fd < 0) {
-		perror(_("Open XML config file"));
-		fprintf(stderr, _("Treating host \"%s\" as a raw hostname\n"),
-			host);
-		return 0;
+	ret = load_xml_conf_file(vpninfo, &xmlfile, &size);
+	if (ret <= 0) {
+		if (ret == 0)
+			fprintf(stderr, _("Treating host \"%s\" as a raw hostname\n"),
+				host);
+		return ret;
 	}
 
-	if (fstat(fd, &st)) {
-		perror(_("fstat XML config file"));
-		close(fd);
-		return -1;
-	}
-
-	xmlfile = mmap(NULL, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
-	if (xmlfile == MAP_FAILED) {
-		perror(_("mmap XML config file"));
-		close(fd);
-		return -1;
-	}
-
-	if (openconnect_sha1(sha1, xmlfile, st.st_size)) {
+	if (openconnect_sha1(sha1, xmlfile, size)) {
 		fprintf(stderr, _("Failed to SHA1 existing file\n"));
-		close(fd);
 		return -1;
 	}
 
 	for (i = 0; i < SHA1_SIZE; i++)
-		sprintf(&vpninfo->xmlsha1[i*2], "%02x", sha1[i]);
+		snprintf(&vpninfo->xmlsha1[i*2], 3, "%02x", sha1[i]);
 
-	vpn_progress(vpninfo, PRG_TRACE, _("XML config file SHA1: %s\n"),
+	vpn_progress(vpninfo, PRG_DEBUG, _("XML config file SHA1: %s\n"),
 		     vpninfo->xmlsha1);
 
-	xml_doc = xmlReadMemory(xmlfile, st.st_size, "noname.xml", NULL, 0);
-	munmap(xmlfile, st.st_size);
-	close(fd);
+	xml_doc = xmlReadMemory(xmlfile, size, "noname.xml", NULL, 0);
+
+	free(xmlfile);
+
 	if (!xml_doc) {
 		fprintf(stderr, _("Failed to parse XML config file %s\n"),
 			vpninfo->xmlconfig);
